"""
    The model's parameters module
    =============================

    This module defines the main classes containing the model configuration parameters.
    The parameters are typically specified as :class:`~.params.parameter.Parameter` objects.

    There are seven types of parameters arranged in classes:

    * :class:`ScaleParams` contains the model scale parameters. These parameters are used to scale and
      `nondimentionalize`_ the :class:`~.params.parameter.Parameter` of the other parameters classes according to
      their :attr:`~.params.parameter.Parameter.units` attribute.
    * :class:`AtmosphericParams` contains the atmospheric dynamical parameters.
    * :class:`AtmosphericTemperatureParams` containing the atmosphere's temperature and heat-exchange parameters.
    * :class:`OceanicParams` contains the oceanic dynamical parameters.
    * :class:`OceanicTemperatureParams` contains the ocean's temperature and heat-exchange parameters.
    * :class:`GroundParams` contains the ground dynamical parameters (e.g. orography).
    * :class:`GroundTemperatureParams` contains the ground's temperature and heat-exchange parameters.

    These parameters classes are regrouped into a global structure :class:`QgParams` which also contains

    * spectral modes definition of the model
    * physical constants
    * parameters derived from the ones provided by the user
    * helper functions to initialize and parameterize the model

    This global parameters structure is used by the other modules to construct the model's ordinary differential
    equations.

    Warning
    -------

    If a model's parameter is set to `None`, it is assumed to be disabled.


    ---------------------

    Description of the classes
    --------------------------

    .. _nondimentionalize: https://en.wikipedia.org/wiki/Nondimensionalization
"""

import numpy as np
import pickle
import warnings
from abc import ABC

from qgs.params.parameter import Parameter, ScalingParameter, ParametersArray
from qgs.basis.fourier import contiguous_channel_basis, contiguous_basin_basis
from qgs.basis.fourier import ChannelFourierBasis, BasinFourierBasis

from sympy import simplify, Symbol


# TODO: - store model version in a variable somewhere
#       - force or warn the user to define the aspect ratio n at parameter object instantiation


class Params(ABC):
    """Base class for a model's parameters container.

    Parameters
    ----------
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.
    """

    _name = ""

    def __init__(self, dic=None):

        self.set_params(dic)

    def set_params(self, dic):
        """Set the specified parameters values.

        Parameters
        ----------
        dic: dict(float or Parameter)
            A dictionary with the parameters names and values to be assigned.
        """
        if dic is not None:
            for key, val in zip(dic.keys(), dic.values()):
                if key in self.__dict__.keys():
                    if isinstance(self.__dict__[key], Parameter):
                        if isinstance(val, Parameter):
                            self.__dict__[key] = val
                        else:
                            d = self.__dict__[key].__dict__
                            self.__dict__[key] = Parameter(val,
                                                           input_dimensional=d['_input_dimensional'],
                                                           units=d['_units'],
                                                           description=d['_description'],
                                                           scale_object=d['_scale_object'],
                                                           symbol=d['_symbol'],
                                                           return_dimensional=d['_return_dimensional'])
                    elif isinstance(self.__dict__[key], ScalingParameter):
                        if isinstance(val, ScalingParameter):
                            self.__dict__[key] = val
                        else:
                            d = self.__dict__[key].__dict__
                            self.__dict__[key] = ScalingParameter(val,
                                                                  units=d['_units'],
                                                                  description=d['_description'],
                                                                  symbol=d['_symbol'],
                                                                  dimensional=d['_dimensional'])
                    else:
                        self.__dict__[key] = val

    def __str__(self):
        s = ""
        for key, val in zip(self.__dict__.keys(), self.__dict__.values()):
            if 'params' not in key and key[0] != '_':
                if val is None:
                    pass
                elif isinstance(val, Parameter):
                    if val.input_dimensional:
                        units = val.units
                        efval = val.dimensional_value
                    else:
                        efval = val.nondimensional_value
                        if val.nondimensional_value == val.dimensional_value:
                            units = ""
                        else:
                            units = "[nondim]"
                    s += "'" + key + "': " + str(efval) + "  " + units + "  (" + val.description + "),\n"
                elif isinstance(val, ScalingParameter):
                    if val.dimensional:
                        units = val.units
                    else:
                        units = "[nondim]"
                    s += "'" + key + "': " + str(val) + "  " + units + "  (" + val.description + "),\n"
                elif isinstance(val, (np.ndarray, list, tuple)) and isinstance(val[0], Parameter):
                    for i, v in enumerate(val):
                        if v.input_dimensional:
                            units = v.units
                            efval = v.dimensional_value
                        else:
                            efval = v.nondimensional_value
                            if v.nondimensional_value == v.dimensional_value:
                                units = ""
                            else:
                                units = "[nondim]"
                        s += "'" + key + "["+str(i+1)+"]': " + str(efval) + "  " + units + "  (" + v.description + "),\n"
                else:
                    s += "'"+key+"': "+str(val)+",\n"
        return s

    def _list_params(self):
        return self._name+" Parameters:\n"+self.__str__()

    def print_params(self):
        """Print the parameters contained in the container."""
        print(self._list_params())

    def __repr__(self):
        s = super(Params, self).__repr__()+"\n"+self._list_params()
        return s

    def load_from_file(self, filename, **kwargs):
        """Function to load previously saved Params object with the method :meth:`save_to_file`.

        Parameters
        ----------
        filename: str
            The file name where the Params object was saved.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'rb')
        tmp_dict = pickle.load(f, **kwargs)
        f.close()

        self.__dict__.clear()
        self.__dict__.update(tmp_dict)

    def save_to_file(self, filename, **kwargs):
        """Function to save the Params object to a file with the :mod:`pickle` module.

        Parameters
        ----------
        filename: str
            The file name where to save the Params object.
        kwargs: dict
            Keyword arguments to pass to the :mod:`pickle` module method.
        """
        f = open(filename, 'wb')
        pickle.dump(self.__dict__, f, **kwargs)
        f.close()


class ScaleParams(Params):
    """Class containing the model scales parameters.

    Parameters
    ----------
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.

    Attributes
    ----------
    scale: Parameter
        The characteristic meridional space scale, :math:`L_y = \\pi \\, L`, in meters [:math:`m`].
    f0: Parameter
        Coriolis parameter, in [:math:`s^{-1}`].
    n: Parameter
        Model domain aspect ratio, :math:`n = 2 L_y/L_x` .
    rra: Parameter
        Earth radius, in meters [:math:`m`].
    phi0_npi: Parameter
        Latitude expressed in fraction of :math:`\\pi` .
    deltap: Parameter
        Difference of pressure between the center of the two atmospheric layers, in [:math:`Pa`].
    """
    _name = "Scale"

    def __init__(self, dic=None):

        Params.__init__(self, dic)

        # -----------------------------------------------------------
        # Scale parameters for the ocean and the atmosphere
        # -----------------------------------------------------------

        self.scale = ScalingParameter(5.e6, units='[m]', description="characteristic space scale (L*pi)", dimensional=True)
        self.f0 = ScalingParameter(1.032e-4, units='[s^-1]', description="Coriolis parameter at the middle of the domain",
                                   dimensional=True, symbol=Symbol('f0'))
        self.n = ScalingParameter(1.3e0, dimensional=False, description="aspect ratio (n = 2 L_y / L_x)", symbol=Symbol('n', positive=True))
        self.rra = ScalingParameter(6370.e3, units='[m]', description="earth radius", dimensional=True)
        self.phi0_npi = ScalingParameter(0.25e0, dimensional=False, description="latitude expressed in fraction of pi")
        self.deltap = ScalingParameter(5.e4, units='[Pa]', description='pressure difference between the two atmospheric layers',
                                       dimensional=True)
        self.Ha = ScalingParameter(8500., units='[m]', description="Average height of the 500 hPa pressure level at midlatitude",
                                   dimensional=True, symbol=Symbol('H_a'))
        self.set_params(dic)

    # ----------------------------------------
    # Some derived parameters (Domain, beta)
    # ----------------------------------------

    @property
    def L(self):
        """Parameter: Typical length scale :math:`L`  of the model, in meters [:math:`m`]."""
        return ScalingParameter(self.scale / np.pi, units=self.scale.units, description='Typical length scale L',
                                symbol=Symbol('L'), dimensional=True)

    @property
    def L_y(self):
        """Parameter: The meridional extent :math:`L_y = \\pi \\, L` of the model's domain, in meters [:math:`m`]."""
        return ScalingParameter(self.scale, units=self.scale.units, description='The meridional extent of the model domain',
                                dimensional=True)

    @property
    def L_x(self):
        """Parameter: The zonal extent :math:`L_x = 2 \\pi \\, L / n` of the model's domain, in meters [:math:`m`]."""
        return ScalingParameter(2 * self.scale / self.n, units=self.scale.units,
                                description='The zonal extent of the model domain',
                                dimensional=True)

    @property
    def phi0(self):
        """Parameter: The reference latitude :math:`\\phi_0` at the center of the domain, expressed in radians [:math:`rad`]."""
        return ScalingParameter(self.phi0_npi * np.pi, units='[rad]',
                                description="The reference latitude of the center of the domain",
                                dimensional=True, symbol=Symbol('phi0'))

    @property
    def beta(self):
        """Parameter: The meridional gradient of the Coriolis parameter at :math:`\\phi_0`, expressed in [:math:`m^{-1} s^{-1}`]. """
        return Parameter(self.L / self.rra * np.cos(self.phi0) / np.sin(self.phi0), input_dimensional=False,
                         units='[m^-1][s^-1]', scale_object=self,
                         description="Meridional gradient of the Coriolis parameter at phi_0", symbol=Symbol('beta'))


class AtmosphericParams(Params):
    """Class containing the atmospheric parameters.

    Parameters
    ----------
    scale_params: ScaleParams
        The scale parameters object of the model.
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.

    Attributes
    ----------
    kd: Parameter
        Atmosphere bottom friction coefficient [:math:`s^{-1}`].
    kdp: Parameter
        Atmosphere internal friction coefficient [:math:`s^{-1}`].
    sigma: Parameter
        Static stability of the atmosphere [:math:`[m^2 s^{-2} Pa^{-2}`].
    """

    _name = "Atmospheric"

    def __init__(self, scale_params, dic=None):

        Params.__init__(self, dic)

        self._scale_params = scale_params

        # Parameters for the atmosphere
        self.kd = Parameter(0.1, input_dimensional=False, scale_object=scale_params, units='[s^-1]',
                            description="atmosphere bottom friction coefficient", symbol=Symbol('k_d'))
        self.kdp = Parameter(0.01, input_dimensional=False, scale_object=scale_params, units='[s^-1]',
                             description="atmosphere internal friction coefficient", symbol=Symbol('k_p'))
        self.sigma = Parameter(0.2e0, input_dimensional=False, scale_object=scale_params, units='[m^2][s^-2][Pa^-2]',
                               description="static stability of the atmosphere", symbol=Symbol('sigma'))

        self.set_params(dic)

    @property
    def sig0(self):
        """Parameter: Static stability of the atmosphere divided by 2."""
        return Parameter(self.sigma / 2, input_dimensional=False, scale_object=self._scale_params, units='[m^2][s^-2][Pa^-2]',
                         description="0.5 * static stability of the atmosphere", symbol=self.sigma.symbol / 2)


class AtmosphericTemperatureParams(Params):
    """Class containing the atmospheric temperature parameters.

    Parameters
    ----------
    scale_params: ScaleParams
        The scale parameters object of the model.
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.

    Attributes
    ----------
    hd: None or Parameter
        Newtonian cooling coefficient.
        Newtonian cooling is disabled if `None`.
    thetas: None or ~numpy.ndarray(float)
        Coefficients of the Newtonian cooling spectral decomposition (non-dimensional).
        Newtonian cooling is disabled if `None`.
    gamma: None or Parameter
        Specific heat capacity of the atmosphere [:math:`J m^{-2} K^{-1}`].
        Heat exchange scheme is disabled if `None`.
    C: None or ~numpy.ndarray(Parameter)
        Spectral decomposition of the constant short-wave radiation of the atmosphere [:math:`W m^{-2}`].
        Heat exchange scheme is disabled if `None`.
    eps: None or Parameter
        Emissivity coefficient for the grey-body atmosphere
        Heat exchange scheme is disabled if `None`.
    T0: None or Parameter
        Stationary solution for the 0-th order atmospheric temperature [:math:`K`].
        Heat exchange scheme is disabled if `None`.
    sc: None or Parameter
        Ratio of surface to atmosphere temperature
        Heat exchange scheme is disabled if `None`.
    hlambda: None or Parameter
        Sensible + turbulent heat exchange between ocean/ground and atmosphere [:math:`W m^{-2} K^{-1}`].
        Heat exchange scheme is disabled if `None`.
    """
    _name = "Atmospheric Temperature"

    def __init__(self, scale_params, dic=None):

        Params.__init__(self, dic)

        self._scale_params = scale_params

        self.hd = Parameter(0.045, input_dimensional=False, units='[s]', scale_object=scale_params,
                            description="Newtonian cooling coefficient", symbol=Symbol('hd'))
        self.thetas = None  # Radiative equilibrium mean temperature decomposition on the model's modes

        self.gamma = None
        self.C = None
        self.eps = None
        self.T0 = None
        self.sc = None
        self.hlambda = None
        self.dynamic_T = None

        self.set_params(dic)

    def set_insolation(self, value, pos=None, dynamic_T=False):
        """Function to define the spectral decomposition of the constant short-wave radiation of the atmosphere (insolation)
        :math:`C_{{\\rm a}, i}` (:attr:`~.AtmosphericTemperatureParams.C`).

        Parameters
        ----------
        value: float, int or iterable
            Value to set. If a scalar is given, the `pos` parameter should be provided to indicate which component to set.
            If an iterable is provided, create a vector of spectral decomposition parameters corresponding to it.
        pos: int, optional
            Indicate in which component to set the `value`.
        dynamic_T: bool, optional
            Whether or not the dynamic temperature scheme is activated.
        """

        # TODO: - check for the dimensionality of the arguments

        if isinstance(value, (float, int)) and pos is not None and self.C is not None:
            offset = 1
            if self.dynamic_T or dynamic_T:
                offset = 0
            self.C[pos] = Parameter(value, units='[W][m^-2]', scale_object=self._scale_params,
                                    description="spectral component "+str(pos+offset)+" of the short-wave radiation of the atmosphere",
                                    return_dimensional=True, symbol=Symbol('C_a'+str(pos+offset)))
        elif hasattr(value, "__iter__"):
            self._create_insolation(value, dynamic_T)
        else:
            warnings.warn('A scalar value was provided, but without the `pos` argument indicating in which ' +
                          'component of the spectral decomposition to put it: Spectral decomposition unchanged !' +
                          'Please specify it or give a vector as `value`.')

    def _create_insolation(self, values, dynamic_T=False):

        if hasattr(values, "__iter__"):
            dim = len(values)
            values = list(values)
        else:
            dim = values
            values = dim * [0.]

        offset = 1
        if dynamic_T:
            offset = 0
            self.dynamic_T = True
        d = ["spectral component "+str(pos+offset)+" of the short-wave radiation of the atmosphere" for pos in range(dim)]
        sy = [Symbol('C_a'+str(pos+offset)) for pos in range(dim)]

        self.C = ParametersArray(values, units='[W][m^-2]', scale_object=self._scale_params,
                                 description=d, return_dimensional=True, symbols=sy)

    def set_thetas(self, value, pos=None):
        """Function to define the spectral decomposition of the Newtonian cooling
        :math:`\\theta^\\star` (:attr:`~.AtmosphericTemperatureParams.thetas`).

        Parameters
        ----------
        value: float, int or iterable
            Value to set. If a scalar is given, the `pos` parameter should be provided to indicate which component to set.
            If an iterable is provided, create a vector of spectral decomposition parameters corresponding to it.
        pos: int, optional
            Indicate in which component to set the `value`.
        """

        # TODO: - check for the dimensionality of the arguments

        if isinstance(value, (float, int)) and pos is not None and self.thetas is not None:
            self.thetas[pos] = Parameter(value, scale_object=self._scale_params,
                                         description="spectral components "+str(pos+1)+" of the temperature profile",
                                         return_dimensional=False, input_dimensional=False, symbol=Symbol('thetas_'+str(pos+1)))
        elif hasattr(value, "__iter__"):
            self._create_thetas(value)
        else:
            warnings.warn('A scalar value was provided, but without the `pos` argument indicating in which ' +
                          'component of the spectral decomposition to put it: Spectral decomposition unchanged !' +
                          'Please specify it or give a vector as `value`.')

    def _create_thetas(self, values):

        if hasattr(values, "__iter__"):
            dim = len(values)
            values = list(values)
        else:
            dim = values
            values = dim * [0.]

        d = ["spectral component "+str(pos+1)+" of the temperature profile" for pos in range(dim)]
        sy = [Symbol('thetas_'+str(pos+1)) for pos in range(dim)]

        self.thetas = ParametersArray(values, scale_object=self._scale_params,
                                      description=d, return_dimensional=False, input_dimensional=False, symbols=sy)


class OceanicParams(Params):
    """Class containing the oceanic parameters

    Parameters
    ----------
    scale_params: ScaleParams
        The scale parameters object of the model.
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.

    Attributes
    ----------
    gp: Parameter
        Reduced gravity in [:math:`m \\, s^{-2}`].
    r: Parameter
        Friction coefficient at the bottom of the ocean in [:math:`s^{-1}`].
    h: Parameter
        Depth of the water layer of the ocean, in meters [:math:`m`].
    d: Parameter
        The strength of the ocean-atmosphere mechanical coupling in [:math:`s^{-1}`].
    """
    _name = "Oceanic"

    def __init__(self, scale_params, dic=None):

        Params.__init__(self, dic)

        self._scale_params = scale_params

        self.gp = Parameter(3.1e-2, units='[m][s^-2]', return_dimensional=True, scale_object=scale_params,
                            description='reduced gravity', symbol=Symbol('g_p'))
        self.r = Parameter(1.e-8, units='[s^-1]', scale_object=scale_params,
                           description="frictional coefficient at the bottom of the ocean", symbol=Symbol('r'))
        self.h = Parameter(5.e2, units='[m]', return_dimensional=True, scale_object=scale_params,
                           description="depth of the water layer of the ocean", symbol=Symbol('h'))
        self.d = Parameter(1.e-8, units='[s^-1]', scale_object=scale_params,
                           description="strength of the ocean-atmosphere mechanical coupling", symbol=Symbol('d'))

        self.set_params(dic)


class OceanicTemperatureParams(Params):
    """Class containing the oceanic temperature parameters

    Parameters
    ----------
    scale_params: ScaleParams
        The scale parameters object of the model.
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.

    Attributes
    ----------
    gamma: None or Parameter
        Specific heat capacity of the ocean [:math:`J m^{-2} K^{-1}`].
        Heat exchange scheme is disabled if `None`.
    C: None or ~numpy.ndarray(Parameter)
        Spectral Decomposition of the constant short-wave radiation of the ocean [:math:`W m^{-2}`].
        Heat exchange scheme is disabled if `None`.
    T0: None or Parameter
        Stationary solution for the 0-th order oceanic temperature [:math:`K`].
        Heat exchange scheme is disabled if `None`.
    """

    _name = "Oceanic Temperature"

    def __init__(self, scale_params, dic=None):

        Params.__init__(self, dic)

        self._scale_params = scale_params

        self.gamma = Parameter(2.e8, units='[J][m^-2][K^-1]', scale_object=scale_params, return_dimensional=True,
                               description='specific heat capacity of the ocean', symbol=Symbol('gamma_o'))
        self.C = None

        self.T0 = None
        self.dynamic_T = None

        self.set_params(dic)

    def set_insolation(self, value, pos=None, dynamic_T=False):
        """Function to define the spectral decomposition of the constant short-wave radiation of the ocean (insolation)
        :math:`C_{{\\rm o}, i}` (:attr:`~.OceanicTemperatureParams.C`).

        Parameters
        ----------
        value: float, int or iterable
            Value to set. If a scalar is given, the `pos` parameter should be provided to indicate which component to set.
            If an iterable is provided, create a vector of spectral decomposition parameters corresponding to it.
        pos: int, optional
            Indicate in which component to set the `value`.
        dynamic_T: bool, optional
            Whether or not the dynamic temperature scheme is activated.
        """

        if isinstance(value, (float, int)) and pos is not None and self.C is not None:
            offset = 1
            if self.dynamic_T or dynamic_T:
                offset = 0
            self.C[pos] = Parameter(value, units='[W][m^-2]', scale_object=self._scale_params,
                                    description="spectral component "+str(pos+offset)+" of the short-wave radiation of the ocean",
                                    return_dimensional=True, symbol=Symbol('C_go'+str(pos+offset)))
        elif hasattr(value, "__iter__"):
            self._create_insolation(value, dynamic_T)
        else:
            warnings.warn('A scalar value was provided, but without the `pos` argument indicating in which ' +
                          'component of the spectral decomposition to put it: Spectral decomposition unchanged !' +
                          'Please specify it or give a vector as `value`.')

    def _create_insolation(self, values, dynamic_T=False):

        if hasattr(values, "__iter__"):
            dim = len(values)
            values = list(values)
        else:
            dim = values
            values = dim * [0.]

        offset = 1
        if dynamic_T:
            offset = 0
            self.dynamic_T = True
        d = ["spectral component "+str(pos+offset)+" of the short-wave radiation of the ocean" for pos in range(dim)]
        sy = [Symbol('C_go'+str(pos+offset)) for pos in range(dim)]

        self.C = ParametersArray(values, units='[W][m^-2]', scale_object=self._scale_params,
                                 description=d, return_dimensional=True, symbols=sy)


class GroundParams(Params):
    """Class containing the ground parameters

    Parameters
    ----------
    scale_params: ScaleParams
        The scale parameters object of the model.
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.

    Attributes
    ----------
    hk: None or ~numpy.ndarray(float)
        Orography spectral decomposition coefficients (non-dimensional), an array of shape (:attr:`~QgParams.nmod` [0],).
        Orography is disabled (flat) if `None`.
    orographic_basis: str
        String to select which component basis modes to use to develop the orography in series.
        Can be either 'atmospheric' or 'ground'. Default to 'atmospheric'.
    """
    _name = "Ground"

    def __init__(self, scale_params, dic=None):

        Params.__init__(self, dic)

        self._scale_params = scale_params

        self.hk = None  # spectral orography coefficients

        self.orographic_basis = "atmospheric"

        self.set_params(dic)

    def set_orography(self, value, pos=None, basis="atmospheric"):
        """Function to define the spectral decomposition of the orography profile
        :math:`h_k` (:attr:`~.GroundParams.hk`).

        Parameters
        ----------
        value: float, int or iterable
            Value to set. If a scalar is given, the `pos` parameter should be provided to indicate which component to set.
            If an iterable is provided, create a vector of spectral decomposition parameters corresponding to it.
        pos: int, optional
            Indicate in which component to set the `value`.
        basis: str, optional
            Indicate which basis should be used to decompose the orography. Can be either `atmospheric`, `oceanic` or `ground`.
            Default to `atmospheric`.
        """

        # TODO: - check for the dimensionality of the arguments
        #       - check that inner products are symbolic if basis is not 'atmospheric'

        self.orographic_basis = basis

        if isinstance(value, (float, int)) and pos is not None and self.hk is not None:
            self.hk[pos] = Parameter(value, scale_object=self._scale_params,
                                     description="spectral components "+str(pos+1)+" of the orography",
                                     return_dimensional=False, input_dimensional=False, symbol=Symbol('hk_'+str(pos+1)))
        elif hasattr(value, "__iter__"):
            self._create_orography(value)
        else:
            warnings.warn('A scalar value was provided, but without the `pos` argument indicating in which ' +
                          'component of the spectral decomposition to put it: Spectral decomposition unchanged !' +
                          'Please specify it or give a vector as `value`.')

    def _create_orography(self, values):

        if hasattr(values, "__iter__"):
            dim = len(values)
            values = list(values)
        else:
            dim = values
            values = dim * [0.]

        d = ["spectral component "+str(pos+1)+" of the orography" for pos in range(dim)]
        sy = [Symbol('hk_'+str(pos+1)) for pos in range(dim)]

        self.hk = ParametersArray(values, scale_object=self._scale_params,
                                  description=d, return_dimensional=False, input_dimensional=False, symbols=sy)


class GroundTemperatureParams(Params):
    """Class containing the ground temperature parameters

    Parameters
    ----------
    scale_params: ScaleParams
        The scale parameters object of the model.
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.

    Attributes
    ----------
    gamma: None or Parameter
        Specific heat capacity of the ground [:math:`J m^{-2} K^{-1}`].
        Heat exchange scheme is disabled if `None`.
    C: None or ~numpy.ndarray(Parameter)
        Spectral decomposition of the constant short-wave radiation of the ground [:math:`W m^{-2}`].
        Heat exchange scheme is disabled if `None`.
    T0: None or Parameter
        Stationary solution for the 0-th order ground temperature [:math:`K`].
        Heat exchange scheme is disabled if `None`.
    """

    _name = "Ground Temperature"

    def __init__(self, scale_params, dic=None):

        Params.__init__(self, dic)

        self._scale_params = scale_params

        self.gamma = Parameter(2.e8, units='[J][m^-2][K^-1]', scale_object=scale_params, return_dimensional=True,
                               description='specific heat capacity of the ground', symbol=Symbol('gamma_g'))
        self.C = None

        self.T0 = None
        self.dynamic_T = None

        self.set_params(dic)

    def set_insolation(self, value, pos=None, dynamic_T=False):
        """Function to define the decomposition of the constant short-wave radiation of the ground (insolation)
        :math:`C_{{\\rm g}, i}` (:attr:`~.GroundTemperatureParams.C`).

        Parameters
        ----------
        value: float, int or iterable
            Value to set. If a scalar is given, the `pos` parameter should be provided to indicate which component to set.
            If an iterable is provided, create a vector of spectral decomposition parameters corresponding to it.
        pos: int, optional
            Indicate in which component to set the `value`.
        dynamic_T: bool, optional
            Whether or not the dynamic temperature scheme is activated.
        """

        # TODO: - check for the dimensionality of the arguments

        if isinstance(value, (float, int)) and pos is not None and self.C is not None:
            offset = 1
            if self.dynamic_T or dynamic_T:
                offset = 0
            self.C[pos] = Parameter(value, units='[W][m^-2]', scale_object=self._scale_params,
                                    description="spectral component "+str(pos+offset)+" of the short-wave radiation of the ground",
                                    return_dimensional=True, symbol=Symbol('C_go'+str(pos+offset)))
        elif hasattr(value, "__iter__"):
            self._create_insolation(value, dynamic_T)
        else:
            warnings.warn('A scalar value was provided, but without the `pos` argument indicating in which ' +
                          'component of the spectral decomposition to put it: Spectral decomposition unchanged !' +
                          'Please specify it or give a vector as `value`.')

    def _create_insolation(self, values, dynamic_T=False):

        if hasattr(values, "__iter__"):
            dim = len(values)
            values = list(values)
        else:
            dim = values
            values = dim * [0.]

        offset = 1
        if dynamic_T:
            offset = 0
            self.dynamic_T = True
        d = ["spectral component "+str(pos+offset)+" of the short-wave radiation of the ground" for pos in range(dim)]
        sy = [Symbol('C_go'+str(pos+offset)) for pos in range(dim)]

        self.C = ParametersArray(values, units='[W][m^-2]', scale_object=self._scale_params,
                                 description=d, return_dimensional=True, symbols=sy)


class QgParams(Params):
    """General qgs parameters container.

    Parameters
    ----------
    dic: dict(float or Parameter), optional
        A dictionary with the parameters names and values to be assigned.
    scale_params: None or ScaleParams, optional
        Scale parameters instance.
        If `None`, create a new ScaleParams instance. Default to None.
        Default to `None`.
    atmospheric_params: bool, None or AtmosphericParams, optional
        Atmospheric parameters instance.
        If 'True`, create a new AtmosphericParams instance.
        If `None`, atmospheric parameters are disabled.
        Default to `True`.
    atemperature_params: bool, None or AtmosphericTemperatureParams, optional
        Atmospheric temperature parameters instance.
        If 'True`, create a new AtmosphericTemperatureParams instance.
        If `None`, atmospheric temperature parameters are disabled.
        Default to `True`.
    oceanic_params: bool, None or OceanicParams, optional
        Oceanic parameters instance.
        If 'True`, create a new OceanicParams instance.
        If `None`, oceanic parameters are disabled.
        Default to `None`.
    otemperature_params: bool, None or OceanicTemperatureParams, optional
        Oceanic temperature parameters instance.
        If 'True`, create a new OceanicTemperatureParams instance.
        If `None`, oceanic temperature parameters are disabled.
        Default to `None`.
    ground_params: bool, None or GroundParams, optional
        Ground parameters instance.
        If 'True`, create a new GroundParams instance.
        If `None`, ground parameters are disabled.
        Default to `True`.
    gtemperature_params: bool, None or GroundTemperatureParams, optional
        Ground temperature parameters instance.
        If 'True`, create a new GroundTemperatureParams instance.
        If `None`, ground temperature parameters are disabled.
        Default to `None`.
    dynamic_T: bool, optional
        Whether to use a fixed or a dynamical reference temperature if the heat exchange scheme is activated.
        Default to `False`.
    T4: bool, optional
        Use or not the :math:`T^4` forcing for the evolution of the temperature field if the heat exchange is activated.
        Activate also the dynamical 0-th temperature.
        Default to `False`.

    Attributes
    ----------
    scale_params: ScaleParams
        Scale parameters instance.
    atmospheric_params: None or AtmosphericParams
        Atmospheric parameters instance.
        If `None`, atmospheric parameters are disabled.
    atemperature_params: None or AtmosphericTemperatureParams
        Atmospheric temperature parameters instance.
        If `None`, atmospheric temperature parameters are disabled.
    oceanic_params: None or OceanicParams
        Oceanic parameters instance.
        If `None`, oceanic parameters are disabled.
    ground_params: None or GroundParams
        Ground parameters instance
        If `None`, ground parameters are disabled.
    gotemperature_params: None, OceanicTemperatureParams or GroundTemperatureParams
        Ground or Oceanic temperature parameters instance.
        If `None`, ground and oceanic temperature parameters are disabled.
    time_unit: float
        Dimensional unit of time to be used to represent the data.
    rr: Parameter
        `Gas constant`_ of `dry air`_ in [:math:`J \\, kg^{-1} \\, K^{-1}`].
    sb: float
        `Stefan-Boltzmann constant`_ in [:math:`J \\, m^{-2} \\, s^{-1} \\, K^{-4}`].
    dynamic_T: bool
        Whether to use a fixed or a dynamical reference temperature if the heat exchange scheme is activated.
        The atmospheric and possibly oceanic (or ground) basis must be reset if this parameter is changed.
    T4: bool
        Use or not the :math:`T^4` forcing for the evolution of the temperature field if the heat exchange is activated.


    .. _Gas constant: https://en.wikipedia.org/wiki/Gas_constant
    .. _dry air: https://en.wikipedia.org/wiki/Gas_constant#Specific_gas_constant
    .. _Stefan-Boltzmann constant: https://en.wikipedia.org/wiki/Stefan%E2%80%93Boltzmann_constant

    """
    _name = "General"

    def __init__(self, dic=None, scale_params=None,
                 atmospheric_params=True, atemperature_params=True,
                 oceanic_params=None, otemperature_params=None,
                 ground_params=True, gtemperature_params=None,
                 dynamic_T=False, T4=False):

        Params.__init__(self, dic)

        # General scale parameters object (Mandatory param block)
        if scale_params is None:
            self.scale_params = ScaleParams(dic)
        else:
            self.scale_params = scale_params

        # Atmospheric parameters object
        if atmospheric_params is True:
            self.atmospheric_params = AtmosphericParams(self.scale_params, dic=dic)
        else:
            self.atmospheric_params = atmospheric_params

        # Atmospheric temperature parameters object
        if atmospheric_params is True:
            self.atemperature_params = AtmosphericTemperatureParams(self.scale_params, dic=dic)
        else:
            self.atemperature_params = atemperature_params

        if oceanic_params is True:
            self.oceanic_params = OceanicParams(self.scale_params, dic)
        else:
            self.oceanic_params = oceanic_params

        if ground_params is True:
            self.ground_params = GroundParams(self.scale_params, dic)
        else:
            self.ground_params = ground_params

        if otemperature_params is True:
            self.gotemperature_params = OceanicTemperatureParams(self.scale_params, dic)
        else:
            self.gotemperature_params = otemperature_params

        if gtemperature_params is True:
            self.gotemperature_params = GroundTemperatureParams(self.scale_params, dic)
        else:
            self.gotemperature_params = gtemperature_params

        self._atmospheric_basis = None
        self._oceanic_basis = None
        self._ground_basis = None

        self._number_of_atmospheric_modes = 0
        self._number_of_oceanic_modes = 0
        self._number_of_ground_modes = 0
        self._ams = None
        self._oms = None
        self._gms = None

        self.dynamic_T = dynamic_T
        self.T4 = T4
        # Force dynamic temperatures if T4 tendencies are activated
        if T4:
            self.dynamic_T = T4

        self._atmospheric_latex_var_string = list()
        self._atmospheric_var_string = list()
        self._oceanic_latex_var_string = list()
        self._oceanic_var_string = list()
        self._ground_latex_var_string = list()
        self._ground_var_string = list()
        self._components_units = [r'm$^2$s$^{-1}$', r'K', r'm$^2$s$^{-1}$', r'K']
        self.time_unit = 'days'

        # Physical constants
        self.rr = Parameter(287.058e0, return_dimensional=True, units='[J][kg^-1][K^-1]',
                            scale_object=self.scale_params, description="gas constant of dry air", symbol=Symbol('R'))
        self.sb = Parameter(5.67e-8, return_dimensional=True, units='[J][m^-2][s^-1][K^-4]',
                            scale_object=self.scale_params, description="Stefan-Boltzmann constant", symbol=Symbol('sigma_b'))

        self.set_params(dic)

    # -----------------------------------------------------------
    # Derived Quantities (Parameters)
    # -----------------------------------------------------------

    @property
    def LR(self):
        """float: Reduced Rossby deformation radius :math:`L_\\text{R} = \\sqrt{g' \\, h } / f_0` ."""
        op = self.oceanic_params
        scp = self.scale_params
        if op is not None:
            try:
                return (op.gp * op.h) ** 0.5 / scp.f0
            except:
                return None
        else:
            return None

    @property
    def G(self):
        """float: The :math:`G = - L^2/L_R^2` parameter."""
        scp = self.scale_params
        if self.LR is not None:
            try:
                return -scp.L**2 / self.LR**2
            except:
                return None
        else:
            return None

    @property
    def Cpgo(self):
        """float: The :math:`C\'_{{\\rm g/\\rm o},i} = R C_{{\\rm g/\\rm o},i} / (\\gamma_{\\rm g/\\rm o} L^2 f_0^3)` parameter."""
        gotp = self.gotemperature_params
        scp = self.scale_params
        if gotp is not None:
            try:
                return gotp.C / (gotp.gamma * scp.f0) * self.rr / (scp.f0 ** 2 * scp.L ** 2)
            except:
                return None
        else:
            return None

    @property
    def Lpgo(self):
        """float: The :math:`\\lambda\'_{{\\rm g/\\rm o}} = \\lambda/(\\gamma_{\\rm g/\\rm o} f_0)` parameter."""
        atp = self.atemperature_params
        gotp = self.gotemperature_params
        scp = self.scale_params
        if atp is not None and gotp is not None:
            try:
                return atp.hlambda / (gotp.gamma * scp.f0)
            except:
                return None
        else:
            return None

    @property
    def Cpa(self):
        """float: The :math:`C\'_{{\\rm a},i} = R C_{{\\rm a},i} / (2 \\gamma_{\\rm a} L^2 f_0^3)` parameter."""
        atp = self.atemperature_params
        scp = self.scale_params
        if atp is not None:
            try:
                return atp.C / (atp.gamma * scp.f0) * self.rr / (scp.f0 ** 2 * scp.L ** 2) / 2
            except:
                return None
        else:
            return None

    @property
    def Lpa(self):
        """float: The :math:`\\lambda\'_{\\rm a} = \\lambda / (\\gamma_{\\rm a} f_0)` parameter."""
        atp = self.atemperature_params
        scp = self.scale_params
        if atp is not None:
            try:
                return atp.hlambda / (atp.gamma * scp.f0)
            except:
                return None
        else:
            return None

    @property
    def sbpgo(self):
        """float: Long wave radiation lost by ground/ocean to the atmosphere :math:`s_{B,{\\rm g/\\rm o}} = 4\\,\\sigma_B \\, T_{{\\rm a},0}^3 / (\\gamma_{\\rm g/\\rm o} f_0)` in the linearized temperature model equations."""
        gotp = self.gotemperature_params
        scp = self.scale_params
        if gotp is not None and not self.dynamic_T:
            try:
                return 4 * self.sb * gotp.T0 ** 3 / (gotp.gamma * scp.f0)
            except:
                return None
        else:
            return None

    @property
    def sbpa(self):
        """float: Long wave radiation from atmosphere absorbed by ground/ocean :math:`s_{B,{\\rm a}} = 4\\,\\epsilon_{\\rm a}\\, \\sigma_B \\, T_{{\\rm a},0}^3 / (\\gamma_{\\rm g/\\rm o} f_0)` in the linearized temperature model equations."""
        atp = self.atemperature_params
        gotp = self.gotemperature_params
        scp = self.scale_params
        if gotp is not None and atp is not None and not self.dynamic_T:
            try:
                return 8 * atp.eps * self.sb * atp.T0 ** 3 / (gotp.gamma * scp.f0)
            except:
                return None
        else:
            return None

    @property
    def LSBpgo(self):
        """float: Long wave radiation from ground/ocean absorbed by atmosphere :math:`S_{B,{\\rm g/\\rm o}} = 2\\,\\epsilon_{\\rm a}\\, \\sigma_B \\, T_{{\\rm a},0}^3 / (\\gamma_{\\rm a} f_0)` in the linearized temperature model equations."""
        atp = self.atemperature_params
        gotp = self.gotemperature_params
        scp = self.scale_params
        if atp is not None and gotp is not None and not self.dynamic_T:
            try:
                return 2 * atp.eps * self.sb * gotp.T0 ** 3 / (atp.gamma * scp.f0)
            except:
                return None
        else:
            return None

    @property
    def LSBpa(self):
        """float: Long wave radiation lost by atmosphere to space & ground/ocean :math:`S_{B,{\\rm a}} = 8\\,\\epsilon_{\\rm a}\\, \\sigma_B \\, T_{{\\rm a},0}^3 / (\\gamma_{\\rm a} f_0)` in the linearized temperature model equations."""
        atp = self.atemperature_params
        scp = self.scale_params
        if atp is not None and not self.dynamic_T:
            try:
                return 8 * atp.eps * self.sb * atp.T0 ** 3 / (atp.gamma * scp.f0)
            except:
                return None
        else:
            return None

    @property
    def T4sbpgo(self):
        """float: Long wave radiation lost by ground/ocean to the atmosphere :math:`s_{B,{\\rm g/\\rm o}} = \\sigma_B \\, L^6 \\, f_0^5 / (\\gamma_{\\rm g/\\rm o} R^3)` in the :math:`T^4` model equations."""
        gotp = self.gotemperature_params
        scp = self.scale_params
        if gotp is not None:
            try:
                return self.sb * scp.L ** 6 * scp.f0 ** 5 / (gotp.gamma * self.rr ** 3)
            except:
                return None
        else:
            return None

    @property
    def T4sbpa(self):
        """float: Long wave radiation from atmosphere absorbed by ground/ocean :math:`s_{B,{\\rm a}} = 16 \\,\\epsilon_{\\rm a}\\, \\sigma_B \\, L^6 \\, f_0^5 / (\\gamma_{\\rm g/\\rm o} R^3)` in the :math:`T^4` model equations."""
        atp = self.atemperature_params
        gotp = self.gotemperature_params
        scp = self.scale_params
        if gotp is not None and atp is not None:
            try:
                return 16 * atp.eps * self.sb * scp.L ** 6 * scp.f0 ** 5 / (gotp.gamma * self.rr ** 3)
            except:
                return None
        else:
            return None

    @property
    def T4LSBpgo(self):
        """float: Long wave radiation from ground/ocean absorbed by atmosphere :math:`S_{B,{\\rm g/\\rm o}} = \\frac{1}{2} \\, \\epsilon_{\\rm a}\\, \\sigma_B \\, L^6 \\, f_0^5 / (\\gamma_{\\rm a} R^3)` in the :math:`T^4` model equations."""
        atp = self.atemperature_params
        scp = self.scale_params
        if atp is not None:
            try:
                return 0.5 * atp.eps * self.sb * scp.L ** 6 * scp.f0 ** 5 / (atp.gamma * self.rr ** 3)
            except:
                return None
        else:
            return None

    @property
    def T4LSBpa(self):
        """float: Long wave radiation lost by atmosphere to space & ground/ocean :math:`S_{B,{\\rm a}} = 16 \\,\\epsilon_{\\rm a}\\, \\sigma_B \\,  L^6 \\, f_0^5 / (\\gamma_{\\rm a} R^3)` in the :math:`T^4` model equations."""
        atp = self.atemperature_params
        scp = self.scale_params
        if atp is not None:
            try:
                return 16 * atp.eps * self.sb * scp.L ** 6 * scp.f0 ** 5 / (atp.gamma * self.rr ** 3)
            except:
                return None
        else:
            return None

    # The following properties might be refactored if the unit system of the model get more widespread across modules.
    @property
    def streamfunction_scaling(self):
        """float: Dimensional scaling of the streamfunction fields."""
        return self.scale_params.L**2 * self.scale_params.f0

    @property
    def temperature_scaling(self):
        """float: Dimensional scaling of the temperature fields."""
        return self.streamfunction_scaling * self.scale_params.f0 / self.rr

    @property
    def geopotential_scaling(self):
        """float: Dimensional scaling of the geopotential height."""
        return self.scale_params.f0 / 9.81

    def set_params(self, dic):
        """Set the specified parameters values.

        Parameters
        ----------
        dic: dict(float or Parameter)
            A dictionary with the parameters names and values to be assigned.
        """
        if dic is not None:
            for key, val in zip(dic.keys(), dic.values()):
                if key in self.__dict__.keys():
                    if isinstance(self.__dict__[key], Parameter):
                        if isinstance(val, Parameter):
                            self.__dict__[key] = val
                        else:
                            d = self.__dict__[key].__dict__
                            self.__dict__[key] = Parameter(val,
                                                           input_dimensional=d['_input_dimensional'],
                                                           units=d['_units'],
                                                           description=d['_description'],
                                                           scale_object=d['_scale_object'],
                                                           symbol=d['_symbol'],
                                                           return_dimensional=d['_return_dimensional'])
                    else:
                        self.__dict__[key] = val

            if 'scale_params' in self.__dict__.keys():
                self.scale_params.set_params(dic)

            if 'atmospheric_params' in self.__dict__.keys():
                if self.atmospheric_params is not None:
                    self.atmospheric_params.set_params(dic)

            if 'atemperature_params' in self.__dict__.keys():
                if self.atemperature_params is not None:
                    self.atemperature_params.set_params(dic)

            if 'oceanic_params' in self.__dict__.keys():
                if self.oceanic_params is not None:
                    self.oceanic_params.set_params(dic)

            if 'ground_params' in self.__dict__.keys():
                if self.ground_params is not None:
                    self.ground_params.set_params(dic)

            if 'otemperature_params' in self.__dict__.keys():
                if self.gotemperature_params is not None:
                    self.gotemperature_params.set_params(dic)

            if 'gtemperature_params' in self.__dict__.keys():
                if self.gotemperature_params is not None:
                    self.gotemperature_params.set_params(dic)

    def print_params(self):
        """Print all the parameters in the container."""
        s = self._list_params()+"\n"
        if 'scale_params' in self.__dict__.keys():
            s += self.scale_params._list_params()+"\n"
        if 'atmospheric_params' in self.__dict__.keys():
            if self.atmospheric_params is not None:
                s += self.atmospheric_params._list_params()+"\n"

        if 'atemperature_params' in self.__dict__.keys():
            if self.atemperature_params is not None:
                s += self.atemperature_params._list_params()+"\n"

        if 'oceanic_params' in self.__dict__.keys():
            if self.oceanic_params is not None:
                s += self.oceanic_params._list_params()+"\n"

        if 'ground_params' in self.__dict__.keys():
            if self.ground_params is not None:
                s += self.ground_params._list_params()+"\n"

        if 'gotemperature_params' in self.__dict__.keys():
            if self.gotemperature_params is not None:
                s += self.gotemperature_params._list_params() + "\n"

        print("Qgs v1.0.0 parameters summary")
        print("=============================\n")
        print(s)

    @property
    def ndim(self):
        """int: Total number of variables of the model."""
        return self.variables_range[-1]

    @property
    def nmod(self):
        """(int, int): Atmospheric and ground/oceanic number of modes."""
        if self._number_of_oceanic_modes != 0:
            return [self._number_of_atmospheric_modes, self._number_of_oceanic_modes]
        else:
            return [self._number_of_atmospheric_modes, self._number_of_ground_modes]

    @property
    def var_string(self):
        """list(str): List of model's variable names."""
        ls = list()
        for var in self._atmospheric_var_string+self._oceanic_var_string+self._ground_var_string:
            ls.append(var)

        return ls

    @property
    def latex_var_string(self):
        """list(str): List of model's variable names, ready for use in latex."""
        ls = list()
        for var in self._atmospheric_latex_var_string+self._oceanic_latex_var_string+self._ground_latex_var_string:
            ls.append(r'{\ '[0:-1] + var + r'}')

        return ls

    @property
    def variables_range(self):
        """list(int): List of the variables indices upper bound per component."""
        natm = self.nmod[0]
        ngoc = self.nmod[1]
        vr = list()
        vr.append(natm)
        vr.append(vr[-1] + natm)
        if self.dynamic_T:
            vr[-1] += 1
        if ngoc > 0:
            vr.append(vr[-1] + ngoc)
            if self._oceanic_basis is not None:
                vr.append(vr[-1] + ngoc)
            if self.dynamic_T:
                vr[-1] += 1
        return vr

    @property
    def number_of_variables(self):
        """list(int): Number of variables per component."""
        vr = self.variables_range
        return [vr[0]] + [vr[i] - vr[i-1] for i in range(1, len(vr))]

    @property
    def latex_components_units(self):
        """list(str): The units of every model's components variables, as a list of latex strings."""
        return self._components_units

    def get_variable_units(self, i):
        """Return the units of a model's variable as a string containing latex symbols.

        Parameters
        ----------
        i: int
            The number of the variable.

        Returns
        -------
        str:
            The string with the units of the variable.
        """
        if i >= self.ndim:
            warnings.warn("Variable " + str(i) + " doesn't exist, cannot return its units.")
            return None
        else:
            if i < self.variables_range[0]:
                return self._components_units[0]
            if self.variables_range[0] <= i < self.variables_range[1]:
                return self._components_units[1]
            if self.oceanic_basis is not None:
                if self.variables_range[1] <= i < self.variables_range[2]:
                    return self._components_units[2]
                if self.variables_range[2] <= i < self.variables_range[3]:
                    return self._components_units[3]
            if self.ground_basis is not None:
                if self.variables_range[1] <= i < self.variables_range[2]:
                    return self._components_units[3]

    @property
    def dimensional_time(self):
        """float: Return the conversion factor between the non-dimensional time and the dimensional time unit specified
        in :attr:`.time_unit`"""
        c = 24 * 3600
        if self.time_unit == 'hours':
            c = 3600
        if self.time_unit == 'days':
            c = 24 * 3600
        if self.time_unit == 'years':
            c = 24 * 3600 * 365
        return 1 / (self.scale_params.f0 * c)

    def _parameter_values(self, obj=None):
        """Function produces a list of the values in the parameters class, or any class passed"""

        subs = list()
        iter_vals = obj.__dict__.values() if obj is not None else self.__dict__.values()
        for val in iter_vals:
            if isinstance(val, Parameter):
                subs.append(val)

            if isinstance(val, ScalingParameter):
                subs.append(val)

            if isinstance(val, ParametersArray):
                for v in val:
                    if v.symbol != 0:
                        subs.append(v)
        return subs

    @property
    def _all_items(self):
        """
        Function to return a list of all values in the parameter class,
        along with their current numerical values.

        Returns
        -------
        list
        """

        subs = self._parameter_values()

        for _, obj in self.__dict__.items():
            if issubclass(obj.__class__, Params):
                for v in self._parameter_values(obj):
                    subs.append(v)

        # Manually add properties from scaling class
        subs.append(self.scale_params.L)
        subs.append(self.scale_params.beta)

        return subs

    # -------------------------------------------------------------------
    # Config setters to be used with symbolic inner products
    # -------------------------------------------------------------------

    @property
    def atmospheric_basis(self):
        """Basis: The atmospheric basis of functions used to project the PDEs onto."""
        return self._atmospheric_basis

    @atmospheric_basis.setter
    def atmospheric_basis(self, basis):
        self._ams = None
        self._oms = None
        self._gms = None

        self._atmospheric_basis = basis
        self._number_of_atmospheric_modes = len(basis.functions)
        if self.dynamic_T:
            self._atmospheric_basis.functions.insert(0, simplify("1"))

        if self.ground_params is not None and self.ground_params.orographic_basis == "atmospheric":
            self.ground_params.set_orography(self._number_of_atmospheric_modes * [0.e0])

        if self.atemperature_params is not None:
            self.atemperature_params.set_thetas(self._number_of_atmospheric_modes * [0.e0])

    @property
    def oceanic_basis(self):
        """Basis: The oceanic basis of functions used to project the PDEs onto."""
        return self._oceanic_basis

    @oceanic_basis.setter
    def oceanic_basis(self, basis):
        self._ams = None
        self._oms = None
        self._gms = None

        self._oceanic_basis = basis
        self._number_of_ground_modes = 0
        self._number_of_oceanic_modes = len(basis)
        if self.dynamic_T:
            self._oceanic_basis.functions.insert(0, simplify("1"))

        if self.atemperature_params is not None:
            # disable the Newtonian cooling
            self.atemperature_params.thetas = None
            self.atemperature_params.hd = None

            self.atemperature_params.gamma = Parameter(1.e7, units='[J][m^-2][K^-1]', scale_object=self.scale_params,
                                                       description='specific heat capacity of the atmosphere',
                                                       return_dimensional=True, symbol=Symbol('gamma_a'))
            if self.dynamic_T:
                self.atemperature_params.set_insolation((self.nmod[0] + 1) * [0.e0], None, True)
                self.atemperature_params.set_insolation(100.0, 0, True)
                self.atemperature_params.set_insolation(100.0, 1, True)
            else:
                self.atemperature_params.set_insolation(self.nmod[0] * [0.e0])
                self.atemperature_params.set_insolation(100.0, 0)
                self.atemperature_params.T0 = Parameter(270.0, units='[K]', scale_object=self.scale_params,
                                                        return_dimensional=True,
                                                        description="stationary solution for the 0-th order atmospheric temperature",
                                                        symbol=Symbol('T_a0'))
            self.atemperature_params.eps = Parameter(0.76e0, input_dimensional=False,
                                                     description="emissivity coefficient for the grey-body atmosphere",
                                                     symbol=Symbol('epsilon'))
            self.atemperature_params.sc = Parameter(1., input_dimensional=False,
                                                    description="ratio of surface to atmosphere temperature",
                                                    symbol=Symbol('sc'))
            self.atemperature_params.hlambda = Parameter(20.00, units='[W][m^-2][K^-1]', scale_object=self.scale_params,
                                                         return_dimensional=True,
                                                         description="sensible+turbulent heat exchange between ocean/ground and atmosphere",
                                                         symbol=Symbol('lambda'))

        if self.gotemperature_params is not None:
            if self.dynamic_T:
                self.gotemperature_params.set_insolation((self.nmod[0] + 1) * [0.e0], None, True)
                self.gotemperature_params.set_insolation(350.0, 0, True)
                self.gotemperature_params.set_insolation(350.0, 1, True)
            else:
                self.gotemperature_params.set_insolation(self.nmod[0] * [0.e0])
                self.gotemperature_params.set_insolation(350.0, 0)
                self.gotemperature_params.T0 = Parameter(285.0, units='[K]', scale_object=self.scale_params, return_dimensional=True,
                                                         description="stationary solution for the 0-th order oceanic temperature",
                                                         symbol=Symbol('T_go0'))
            # if setting an ocean, then disable the orography
            if self.ground_params is not None:
                self.ground_params.hk = None

    @property
    def ground_basis(self):
        """Basis: The ground basis of functions used to project the PDEs onto."""
        return self._ground_basis

    @ground_basis.setter
    def ground_basis(self, basis):
        self._ams = None
        self._oms = None
        self._gms = None

        if basis[0] == 1 or basis[0] == Symbol("1"):
            del basis[0]
        self._ground_basis = basis
        self._number_of_ground_modes = len(basis)
        self._number_of_oceanic_modes = 0

        if self.dynamic_T:
            self._ground_basis.functions.insert(0, simplify("1"))
        if self.atemperature_params is not None:
            # disable the Newtonian cooling
            self.atemperature_params.thetas = None
            self.atemperature_params.hd = None

            self.atemperature_params.gamma = Parameter(1.e7, units='[J][m^-2][K^-1]', scale_object=self.scale_params,
                                                       description='specific heat capacity of the atmosphere',
                                                       return_dimensional=True, symbol=Symbol('gamma_a'))
            if self.dynamic_T:
                self.atemperature_params.set_insolation((self.nmod[0] + 1) * [0.e0], None, True)
                self.atemperature_params.set_insolation(100.0, 0, True)
                self.atemperature_params.set_insolation(100.0, 1, True)
            else:
                self.atemperature_params.set_insolation(self.nmod[0] * [0.e0])
                self.atemperature_params.set_insolation(100.0, 0)
                self.atemperature_params.T0 = Parameter(270.0, units='[K]', scale_object=self.scale_params,
                                                        return_dimensional=True,
                                                        description="stationary solution for the 0-th order atmospheric temperature",
                                                        symbol=Symbol('T_a0'))
            self.atemperature_params.eps = Parameter(0.76e0, input_dimensional=False,
                                                     description="emissivity coefficient for the grey-body atmosphere",
                                                     symbol=Symbol('epsilon'))
            self.atemperature_params.sc = Parameter(1., input_dimensional=False,
                                                    description="ratio of surface to atmosphere temperature",
                                                    symbol=Symbol('sc'))
            self.atemperature_params.hlambda = Parameter(20.00, units='[W][m^-2][K^-1]', scale_object=self.scale_params,
                                                         return_dimensional=True,
                                                         description="sensible+turbulent heat exchange between ocean/ground and atmosphere",
                                                         symbol=Symbol('lambda'))

        if self.gotemperature_params is not None:
            # if orography is disabled, enable it!
            if self.ground_params is not None:
                if self.ground_params.hk is None:
                    if self.ground_params.orographic_basis == 'atmospheric':
                        self.ground_params.set_orography(self._number_of_atmospheric_modes * [0.e0])
                    else:
                        self.ground_params.set_orography(self._number_of_ground_modes * [0.e0])
                    self.ground_params.set_orography(0.1, 1)
            if self.dynamic_T:
                self.gotemperature_params.set_insolation((self.nmod[0] + 1) * [0.e0], None, True)
                self.gotemperature_params.set_insolation(350.0, 0, True)
                self.gotemperature_params.set_insolation(350.0, 1, True)
            else:
                self.gotemperature_params.set_insolation(self.nmod[0] * [0.e0])
                self.gotemperature_params.set_insolation(350.0, 0)
                self.gotemperature_params.T0 = Parameter(285.0, units='[K]', scale_object=self.scale_params, return_dimensional=True,
                                                         description="stationary solution for the 0-th order oceanic temperature",
                                                         symbol=Symbol('T_go0'))

    def set_atmospheric_modes(self, basis, auto=False):
        """Function to configure the atmospheric modes (basis functions) used to project the PDEs onto.

        Parameters
        ----------
        basis: Basis
            Basis object containing the definition of the atmospheric modes.
        auto: bool, optional
            Automatically instantiate the parameters container needed to describe the atmospheric models parameters.
            Default is False.

        Examples
        --------

        >>> from qgs.params.params import QgParams
        >>> from qgs.basis.fourier import contiguous_channel_basis
        >>> q = QgParams()
        >>> atm_basis = contiguous_channel_basis(2, 2, 1.5)
        >>> q.set_atmospheric_modes(atm_basis)
        >>> q.atmospheric_basis
        [sqrt(2)*cos(y), 2*sin(y)*cos(n*x), 2*sin(y)*sin(n*x), sqrt(2)*cos(2*y), 2*sin(2*y)*cos(n*x), 2*sin(2*y)*sin(n*x), 2*sin(y)*cos(2*n*x), 2*sin(y)*sin(2*n*x), 2*sin(2*y)*cos(2*n*x), 2*sin(2*y)*sin(2*n*x)]

        """

        if auto:
            if self.atemperature_params is None:
                self.atemperature_params = AtmosphericTemperatureParams(self.scale_params)
            if self.atmospheric_params is None:
                self.atmospheric_params = AtmosphericParams(self.scale_params)

        self.atmospheric_basis = basis

        self._atmospheric_latex_var_string = list()
        self._atmospheric_var_string = list()
        for i in range(1, self.nmod[0] + 1):
            self._atmospheric_latex_var_string.append(r'psi_{{\rm a},' + str(i) + "}")
            self._atmospheric_var_string.append(r'psi_a_' + str(i))
        if self.dynamic_T:
            self._atmospheric_latex_var_string.append(r', T_{{\rm a},0}')
            self._atmospheric_var_string.append(r'T_a_0')
        for i in range(1, self.nmod[0] + 1):
            self._atmospheric_latex_var_string.append(r'theta_{{\rm a},' + str(i) + "}")
            self._atmospheric_var_string.append(r'theta_a_' + str(i))

    def set_oceanic_modes(self, basis, auto=True):
        """Function to configure the oceanic modes (basis functions) used to project the PDEs onto.

        Parameters
        ----------
        basis: Basis
            Basis object containing the definition of the oceanic modes.
        auto: bool, optional
            Automatically instantiate or not the parameters container needed to describe the oceanic models parameters.
            Default is True.

        Examples
        --------

        >>> from qgs.params.params import QgParams
        >>> from qgs.basis.fourier import contiguous_channel_basis, contiguous_basin_basis
        >>> q = QgParams()
        >>> atm_basis = contiguous_channel_basis(2, 2, 1.5)
        >>> oc_basis = contiguous_basin_basis(2, 4, 1.5)
        >>> q.set_atmospheric_modes(atm_basis)
        >>> q.set_oceanic_modes(oc_basis)
        >>> q.oceanic_basis
        [2*sin(y)*sin(0.5*n*x), 2*sin(2*y)*sin(0.5*n*x), 2*sin(3*y)*sin(0.5*n*x), 2*sin(4*y)*sin(0.5*n*x), 2*sin(y)*sin(1.0*n*x), 2*sin(2*y)*sin(1.0*n*x), 2*sin(3*y)*sin(1.0*n*x), 2*sin(4*y)*sin(1.0*n*x)]

        """
        if self._atmospheric_basis is None:  # Presently, the ocean can not yet be set independently of an atmosphere.
            print('Atmosphere modes not set up. Add an atmosphere before adding an ocean!')
            print('Oceanic setup aborted.')
            return

        if auto:
            if self.gotemperature_params is None or isinstance(self.gotemperature_params, GroundTemperatureParams):
                self.gotemperature_params = OceanicTemperatureParams(self.scale_params)
            if self.oceanic_params is None:
                self.oceanic_params = OceanicParams(self.scale_params)

            self.ground_params = None
            self._ground_basis = None

        self.oceanic_basis = basis

        self._oceanic_latex_var_string = list()
        self._oceanic_var_string = list()
        self._ground_latex_var_string = list()
        self._ground_var_string = list()
        for i in range(1, self.nmod[1] + 1):
            self._oceanic_latex_var_string.append(r'psi_{\rm o,' + str(i) + "}")
            self._oceanic_var_string.append(r'psi_o_' + str(i))
        if self.dynamic_T:
            self._oceanic_latex_var_string.append(r', T_{{\rm o},0}')
            self._oceanic_var_string.append(r'T_o_0')
        for i in range(1, self.nmod[1] + 1):
            self._oceanic_latex_var_string.append(r'delta T_{{\rm o},' + str(i) + "}")
            self._oceanic_var_string.append(r'delta_T_o_' + str(i))

    def set_ground_modes(self, basis=None, auto=True):
        """Function to configure the ground modes (basis functions) used to project the PDEs onto.

        Parameters
        ----------
        basis: None or Basis, optional
            Basis object containing the definition of the ground modes. If `None`, use the basis of the atmosphere.
            Default to `None`.
        auto: bool, optional
            Automatically instantiate or not the parameters container needed to describe the ground models parameters.
            Default is True.

        Examples
        --------

        >>> from qgs.params.params import QgParams
        >>> from qgs.basis.fourier import contiguous_channel_basis, contiguous_basin_basis
        >>> q = QgParams()
        >>> atm_basis = contiguous_channel_basis(2, 2, 1.5)
        >>> q.set_atmospheric_modes(atm_basis)
        >>> q.set_ground_modes()
        >>> q.ground_basis
        [sqrt(2)*cos(y), 2*sin(y)*cos(n*x), 2*sin(y)*sin(n*x), sqrt(2)*cos(2*y), 2*sin(2*y)*cos(n*x), 2*sin(2*y)*sin(n*x), 2*sin(y)*cos(2*n*x), 2*sin(y)*sin(2*n*x), 2*sin(2*y)*cos(2*n*x), 2*sin(2*y)*sin(2*n*x)]
        """
        if self._atmospheric_basis is None:  # Presently, the ground can not yet be set independently of an atmosphere.
            print('Atmosphere modes not set up. Add an atmosphere before adding the ground!')
            print('Ground setup aborted.')
            return

        if auto:
            if self.gotemperature_params is None or isinstance(self.gotemperature_params, OceanicTemperatureParams):
                self.gotemperature_params = GroundTemperatureParams(self.scale_params)
            if self.ground_params is None:
                self.ground_params = GroundParams(self.scale_params)

            self.oceanic_params = None
            self._oceanic_basis = None

        if basis is not None:
            self.ground_basis = basis
        else:
            self.ground_basis = self._atmospheric_basis

        self._oceanic_var_string = list()
        self._oceanic_latex_var_string = list()
        self._ground_latex_var_string = list()
        self._ground_var_string = list()
        if self.dynamic_T:
            self._oceanic_latex_var_string.append(r', T_{{\rm g},0}')
            self._oceanic_var_string.append(r'T_g_0')
        for i in range(1, self.nmod[1] + 1):
            self._ground_latex_var_string.append(r'delta T_{\rm g,' + str(i) + "}")
            self._ground_var_string.append(r'delta_T_g_' + str(i))

    # -------------------------------------------------------------------
    # Specific basis setters
    # -------------------------------------------------------------------

    def set_atmospheric_channel_fourier_modes(self, nxmax, nymax, auto=False, mode='analytic'):
        """Function to configure and set the basis for contiguous spectral blocks of atmospheric modes on a channel.

        Parameters
        ----------
        nxmax: int
            Maximum x-wavenumber to fill the spectral block up to.
        nymax: int
            Maximum :math:`y`-wavenumber to fill the spectral block up to.
        auto: bool, optional
            Automatically instantiate the parameters container needed to describe the atmospheric models parameters.
            Default is `False`.
        mode: str, optional
            Mode to set the inner products: Either `analytic` or `symbolic`:
            `analytic` for inner products computed with formula or `symbolic` using `Sympy`_.
            Default to `analytic`.

        Examples
        --------

        >>> from qgs.params.params import QgParams
        >>> q = QgParams()
        >>> q.set_atmospheric_channel_fourier_modes(2, 2)
        >>> q.ablocks
        array([[1, 1],
               [1, 2],
               [2, 1],
               [2, 2]])

        .. _Sympy: https://www.sympy.org/

        """

        if mode == 'symbolic':
            basis = contiguous_channel_basis(nxmax, nymax, self.scale_params.n)
            self.set_atmospheric_modes(basis, auto)
        else:
            self._set_atmospheric_analytic_fourier_modes(nxmax, nymax, auto)

    def set_oceanic_basin_fourier_modes(self, nxmax, nymax, auto=True, mode='analytic'):
        """Function to configure and set the basis for contiguous spectral blocks of oceanic modes on a closed basin.

        Parameters
        ----------
        nxmax: int
            Maximum x-wavenumber to fill the spectral block up to.
        nymax: int
            Maximum :math:`y`-wavenumber to fill the spectral block up to.
        auto: bool, optional
            Automatically instantiate the parameters container needed to describe the atmospheric models parameters.
            Default is `True`.
        mode: str, optional
            Mode to set the inner products: Either `analytic` or `symbolic`.
            `analytic` for inner products computed with formula or `symbolic` using `Sympy`_.
            Default to `analytic`.

        Examples
        --------

        >>> from qgs.params.params import QgParams
        >>> q = QgParams()
        >>> q.set_atmospheric_channel_fourier_modes(2, 2)
        >>> q.set_oceanic_basin_fourier_modes(2, 4)
        >>> q.oblocks
        array([[1, 1],
               [1, 2],
               [1, 3],
               [1, 4],
               [2, 1],
               [2, 2],
               [2, 3],
               [2, 4]])

        .. _Sympy: https://www.sympy.org/
        """

        if mode == 'symbolic':
            basis = contiguous_basin_basis(nxmax, nymax, self.scale_params.n)
            self.set_oceanic_modes(basis, auto)
        else:
            self._set_oceanic_analytic_fourier_modes(nxmax, nymax, auto)

    def set_ground_channel_fourier_modes(self, nxmax=None, nymax=None, auto=True, mode='analytic'):
        """Function to configure and set the basis for contiguous spectral blocks of ground modes on a channel.

        Parameters
        ----------
        nxmax: int, optional
            Maximum x-wavenumber to fill the spectral block up to. Default to `None`.
        nymax: int, optional
            Maximum :math:`y`-wavenumber to fill the spectral block up to. Default to `None`.
        auto: bool, optional
            Automatically instantiate the parameters container needed to describe the atmospheric models parameters.
            Default is `True`.
        mode: str, optional
            Mode to set the inner products: Either `analytic` or `symbolic`.
            `analytic` for inner products computed with formula or `symbolic` using `Sympy`_.
            Default to `analytic`.

        Notes
        -----
        If both `nxmax` and `nymax` are `None`, default to the atmospheric basis configuration if available.

        Examples
        --------

        >>> from qgs.params.params import QgParams
        >>> q = QgParams()
        >>> q.set_atmospheric_channel_fourier_modes(2,4)
        >>> q.set_ground_channel_fourier_modes()
        >>> q.gblocks
        array([[1, 1],
               [1, 2],
               [1, 3],
               [1, 4],
               [2, 1],
               [2, 2],
               [2, 3],
               [2, 4]])

        .. _Sympy: https://www.sympy.org/
        """

        if mode == "symbolic":
            if nxmax is not None and nymax is not None:
                basis = contiguous_channel_basis(nxmax, nymax, self.scale_params.n)
            else:
                basis = None
            self.set_ground_modes(basis, auto)
        else:
            self._set_ground_analytic_fourier_modes(nxmax, nymax, auto)

    # -------------------------------------------------------------------
    # Model configs setter to be used with analytic inner products
    # -------------------------------------------------------------------

    @property
    def ablocks(self):
        """~numpy.ndarray(int): Spectral blocks detailing the model's atmospheric modes :math:`x`- and :math:`y`-wavenumber.
         Array of shape (:attr:`~QgParams.nmod` [0], 2)."""
        return self._ams

    @ablocks.setter
    def ablocks(self, value):
        self._ams = value
        basis = ChannelFourierBasis(self._ams, self.scale_params.n)
        self._atmospheric_basis = basis

        namod = 0
        for i in range(self.ablocks.shape[0]):
            if self.ablocks[i, 0] == 1:
                namod += 3
            else:
                namod += 2

        self._number_of_atmospheric_modes = namod
        if self.ground_params is not None:
            self.ground_params.orographic_basis = 'atmospheric'
            self.ground_params.set_orography(namod * [0.e0])
            self.ground_params.set_orography(0.1, 1)
        if self.atemperature_params is not None:
            self.atemperature_params.set_thetas(namod * [0.e0])
            self.atemperature_params.set_thetas(0.1, 0)

    @property
    def oblocks(self):
        """~numpy.ndarray(int): Spectral blocks detailing the model's oceanic modes :math:`x`-and :math:`y`-wavenumber.
         Array of shape (:attr:`~QgParams.nmod` [1], 2)."""
        return self._oms

    @oblocks.setter
    def oblocks(self, value):
        self._oms = value
        self._gms = None
        basis = BasinFourierBasis(self._oms, self.scale_params.n)
        self._oceanic_basis = basis
        self._ground_basis = None

        if self.atemperature_params is not None:
            # disable the Newtonian cooling
            self.atemperature_params.thetas = None  # np.zeros(self.nmod[0])
            self.atemperature_params.hd = None  # Parameter(0.0, input_dimensional=False)

            self.atemperature_params.gamma = Parameter(1.e7, units='[J][m^-2][K^-1]', scale_object=self.scale_params,
                                                       description='specific heat capacity of the atmosphere',
                                                       return_dimensional=True,
                                                       symbol=Symbol('gamma_a'))
            self.atemperature_params.set_insolation(self.nmod[0] * [0.e0])
            self.atemperature_params.set_insolation(100.0, 0)
            self.atemperature_params.eps = Parameter(0.76e0, input_dimensional=False,
                                                     description="emissivity coefficient for the grey-body atmosphere",
                                                     symbol=Symbol('epsilon'))
            self.atemperature_params.T0 = Parameter(270.0, units='[K]', scale_object=self.scale_params,
                                                    return_dimensional=True,
                                                    description="stationary solution for the 0-th order atmospheric temperature",
                                                    symbol=Symbol('T_a0'))
            self.atemperature_params.sc = Parameter(1., input_dimensional=False,
                                                    description="ratio of surface to atmosphere temperature",
                                                    symbol=Symbol('sc'))
            self.atemperature_params.hlambda = Parameter(20.00, units='[W][m^-2][K^-1]', scale_object=self.scale_params,
                                                         return_dimensional=True,
                                                         description="sensible+turbulent heat exchange between ocean/ground and atmosphere",
                                                         symbol=Symbol('lambda'))

        if self.gotemperature_params is not None:
            self._number_of_ground_modes = 0
            self._number_of_oceanic_modes = self.oblocks.shape[0]
            self.gotemperature_params.set_insolation(self.nmod[0] * [0.e0])
            self.gotemperature_params.set_insolation(350.0, 0)
            self.gotemperature_params.T0 = Parameter(285.0, units='[K]', scale_object=self.scale_params, return_dimensional=True,
                                                     description="stationary solution for the 0-th order oceanic temperature",
                                                     symbol=Symbol('T_go0'))
            # if setting an ocean, then disable the orography
            if self.ground_params is not None:
                self.ground_params.hk = None

    @property
    def gblocks(self):
        """~numpy.ndarray(int): Spectral blocks detailing the model's ground modes :math:`x`-and :math:`y`-wavenumber.
         Array of shape (:attr:`~QgParams.nmod` [1], 2)."""
        return self._gms

    @gblocks.setter
    def gblocks(self, value):
        self._oms = None
        self._gms = value
        basis = ChannelFourierBasis(self._gms, self.scale_params.n)
        self._oceanic_basis = None
        self._ground_basis = basis

        if self.atemperature_params is not None:
            # disable the Newtonian cooling
            self.atemperature_params.thetas = None  # np.zeros(self.nmod[0])
            self.atemperature_params.hd = None  # Parameter(0.0, input_dimensional=False)

            self.atemperature_params.gamma = Parameter(1.e7, units='[J][m^-2][K^-1]',
                                                       scale_object=self.scale_params,
                                                       description='specific heat capacity of the atmosphere',
                                                       return_dimensional=True,
                                                       symbol=Symbol('gamma_a'))
            self.atemperature_params.set_insolation(self.nmod[0] * [0.e0])
            self.atemperature_params.set_insolation(100.0, 0)
            self.atemperature_params.eps = Parameter(0.76e0, input_dimensional=False,
                                                     description="emissivity coefficient for the grey-body atmosphere",
                                                     symbol=Symbol('epsilon'))
            self.atemperature_params.T0 = Parameter(270.0, units='[K]', scale_object=self.scale_params,
                                                    return_dimensional=True,
                                                    description="stationary solution for the 0-th order atmospheric temperature",
                                                    symbol=Symbol('T_a0'))
            self.atemperature_params.sc = Parameter(1., input_dimensional=False,
                                                    description="ratio of surface to atmosphere temperature",
                                                    symbol=Symbol('sc'))
            self.atemperature_params.hlambda = Parameter(20.00, units='[W][m^-2][K^-1]',
                                                         scale_object=self.scale_params,
                                                         return_dimensional=True,
                                                         description="sensible+turbulent heat exchange between ocean/ground and atmosphere",
                                                         symbol=Symbol('lambda'))

        if self.gotemperature_params is not None:
            gmod = 0
            for i in range(self.gblocks.shape[0]):
                if self.ablocks[i, 0] == 1:
                    gmod += 3
                else:
                    gmod += 2
            self._number_of_ground_modes = gmod
            self._number_of_oceanic_modes = 0
            # if orography is disabled, enable it!
            if self.ground_params is not None:
                self.ground_params.orographic_basis = 'atmospheric'
                if self.ground_params.hk is None:
                    self.ground_params.set_orography(self.nmod[0] * [0.e0])
                    self.ground_params.set_orography(0.1, 1)
            self.gotemperature_params.set_insolation(self.nmod[0] * [0.e0])
            self.gotemperature_params.set_insolation(350.0, 0)
            self.gotemperature_params.T0 = Parameter(285.0, units='[K]', scale_object=self.scale_params, return_dimensional=True,
                                                     description="stationary solution for the 0-th order oceanic temperature",
                                                     symbol=Symbol('T_go0'))

    def _set_atmospheric_analytic_fourier_modes(self, nxmax, nymax, auto=False):

        res = np.zeros((nxmax * nymax, 2), dtype=int)
        i = 0
        for nx in range(1, nxmax + 1):
            for ny in range(1, nymax+1):
                res[i, 0] = nx
                res[i, 1] = ny
                i += 1

        if auto:
            if self.atemperature_params is None:
                self.atemperature_params = AtmosphericTemperatureParams(self.scale_params)
            if self.atmospheric_params is None:
                self.atmospheric_params = AtmosphericParams(self.scale_params)

        self.ablocks = res

        self._atmospheric_latex_var_string = list()
        self._atmospheric_var_string = list()
        for i in range(self.nmod[0]):
            self._atmospheric_latex_var_string.append(r'psi_{\rm a,' + str(i + 1) + "}")
            self._atmospheric_var_string.append(r'psi_a_' + str(i + 1))
        for i in range(self.nmod[0]):
            self._atmospheric_latex_var_string.append(r'theta_{\rm a,' + str(i + 1) + "}")
            self._atmospheric_var_string.append(r'theta_a_' + str(i + 1))

    def _set_oceanic_analytic_fourier_modes(self, nxmax, nymax, auto=True):

        if self._ams is None:
            print('Atmosphere modes not set up. Add an atmosphere before adding an ocean!')
            print('Oceanic setup aborted.')
            return
        res = np.zeros((nxmax * nymax, 2), dtype=int)
        i = 0
        for nx in range(1, nxmax + 1):
            for ny in range(1, nymax+1):
                res[i, 0] = nx
                res[i, 1] = ny
                i += 1

        if auto:
            if self.gotemperature_params is None or isinstance(self.gotemperature_params, GroundTemperatureParams):
                self.gotemperature_params = OceanicTemperatureParams(self.scale_params)
            if self.oceanic_params is None:
                self.oceanic_params = OceanicParams(self.scale_params)

            self.ground_params = None

        self.oblocks = res

        self._oceanic_latex_var_string = list()
        self._oceanic_var_string = list()
        self._ground_latex_var_string = list()
        self._ground_var_string = list()
        for i in range(self.nmod[1]):
            self._oceanic_latex_var_string.append(r'psi_{\rm o,' + str(i + 1) + "}")
            self._oceanic_var_string.append(r'psi_o_' + str(i + 1))
        for i in range(self.nmod[1]):
            self._oceanic_latex_var_string.append(r'delta T_{\rm o,' + str(i + 1) + "}")
            self._oceanic_var_string.append(r'delta_T_o_' + str(i + 1))

    def _set_ground_analytic_fourier_modes(self, nxmax=None, nymax=None, auto=True):

        if self._ams is None:
            print('Atmosphere modes not set up. Add an atmosphere before adding the ground!')
            print('Ground setup aborted.')
            return

        if nxmax is None or nymax is None:
            res = self._ams.copy()
        else:
            res = np.zeros((nxmax * nymax, 2), dtype=int)
            i = 0
            for nx in range(1, nxmax + 1):
                for ny in range(1, nymax+1):
                    res[i, 0] = nx
                    res[i, 1] = ny
                    i += 1

        if auto:
            if self.gotemperature_params is None or isinstance(self.gotemperature_params, OceanicTemperatureParams):
                self.gotemperature_params = GroundTemperatureParams(self.scale_params)
            if self.ground_params is None:
                self.ground_params = GroundParams(self.scale_params)

            self.oceanic_params = None

        self.gblocks = res

        self._oceanic_var_string = list()
        self._oceanic_latex_var_string = list()
        self._ground_latex_var_string = list()
        self._ground_var_string = list()
        for i in range(self.nmod[1]):
            self._ground_latex_var_string.append(r'delta T_{\rm g,' + str(i + 1) + "}")
            self._ground_var_string.append(r'delta_T_g_' + str(i + 1))




"""
    Integrator module
    =================

    Module with the classes of integrators to multi-thread the integration of
    an ordinary differential equations

    .. math:: \dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

    of the model and its linearized version.

    Module classes
    --------------

    * :class:`RungeKuttaIntegrator`
    * :class:`RungeKuttaTglsIntegrator`

"""
import multiprocessing
import numpy as np
from numba import njit
from qgs.integrators.integrate import _integrate_runge_kutta_jit, _integrate_runge_kutta_tgls_jit, _zeros_func
from qgs.functions.util import reverse


class RungeKuttaIntegrator(object):
    """Class to integrate the ordinary differential equations (ODEs)

    .. math:: \dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

    with a set of :class:`TrajectoryProcess` and a specified `Runge-Kutta method`_.

    .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

    Parameters
    ----------
    num_threads: None or int, optional
        Number of :class:`TrajectoryProcess` workers (threads) to use. If `None`, use the number of machine's
        cores available. Default to `None`.
    b: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    c: None or ~numpy.ndarray, optional
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    a: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    number_of_dimensions: None or int, optional
        Allow to hardcode the dynamical system dimension. If `None`, evaluate the dimension from the
        callable :attr:`func`. Default to `None`.

    Attributes
    ----------
    num_threads: int
        Number of :class:`TrajectoryProcess` workers (threads) to use.
    b: ~numpy.ndarray
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
    c: ~numpy.ndarray
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
    a: ~numpy.ndarray
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
    n_dim: int
        Dynamical system dimension.
    n_traj: int
        The number of trajectories (initial conditions) computed at the last integration
        performed by the integrator.
    n_records: int
        The number of saved states of the last integration performed by the integrator.
    ic: ~numpy.ndarray
        Store the integrator initial conditions.
    func: callable
        Last function :math:`\\boldsymbol{f}` used by the integrator to integrate.
    """

    def __init__(self, num_threads=None, b=None, c=None, a=None, number_of_dimensions=None):

        if num_threads is None:
            self.num_threads = multiprocessing.cpu_count()
        else:
            self.num_threads = num_threads

        # Default is RK4
        if a is None and b is None and c is None:
            self.c = np.array([0., 0.5, 0.5, 1.])
            self.b = np.array([1./6, 1./3, 1./3, 1./6])
            self.a = np.zeros((len(self.c), len(self.b)))
            self.a[1, 0] = 0.5
            self.a[2, 1] = 0.5
            self.a[3, 2] = 1.
        else:
            self.a = a
            self.b = b
            self.c = c

        self.ic = None
        self._time = None
        self._recorded_traj = None
        self.n_traj = 0
        self.n_dim = number_of_dimensions
        self.n_records = 0
        self._write_steps = 0
        self._time_direction = 1

        self.func = None

        self._ics_queue = None
        self._traj_queue = None

        self._processes_list = list()

    def terminate(self):
        """Stop the workers (threads) and release the resources of the integrator."""

        for process in self._processes_list:

            process.terminate()
            process.join()

    def start(self):
        """Start or restart the workers (threads) of the integrator.

        Warnings
        --------
        If the integrator was not previously terminated, it will be terminated first in the case
        of a restart.
        """

        self.terminate()

        self._processes_list = list()
        self._ics_queue = multiprocessing.JoinableQueue()
        self._traj_queue = multiprocessing.Queue()

        for i in range(self.num_threads):
            self._processes_list.append(TrajectoryProcess(i, self.func, self.b, self.c, self.a,
                                                          self._ics_queue, self._traj_queue))

        for process in self._processes_list:
            process.daemon = True
            process.start()

    def set_func(self, f, ic_init=True):
        """Set the `Numba`_-jitted function :math:`\\boldsymbol{f}` to integrate.

        .. _Numba: https://numba.pydata.org/

        Parameters
        ----------
        f: callable
            The `Numba`_-jitted function :math:`\\boldsymbol{f}`.
            Should have the signature ``f(t, x)`` where ``x`` is the state value and ``t`` is the time.
        ic_init: bool, optional
            Re-initialize or not the initial conditions of the integrator. Default to `True`.

        Warnings
        --------
        This function restarts the integrator!
        """

        self.func = f
        if ic_init:
            self.ic = None
        self.start()

    def set_bca(self, b=None, c=None, a=None, ic_init=True):
        """Set the coefficients of the `Runge-Kutta method`_ and restart the integrator. s

        .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

        Parameters
        ----------
        b: None or ~numpy.ndarray, optional
            Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
            If `None`, does not reinitialize these coefficients.
        c: None or ~numpy.ndarray, optional
            Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
            If `None`, does not reinitialize these coefficients.
        a: None or ~numpy.ndarray, optional
            Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
            If `None`, does not reinitialize these coefficients.
        ic_init: bool, optional
            Re-initialize or not the initial conditions of the integrator. Default to `True`.

        """

        if a is not None:
            self.a = a
        if b is not None:
            self.b = b
        if c is not None:
            self.c = c
        if ic_init:
            self.ic = None
        self.start()

    def initialize(self, convergence_time, dt, pert_size=0.01, reconvergence_time=None, forward=True,
                   number_of_trajectories=1, ic=None, reconverge=False):
        """Initialize the integration on an attractor by running it for a transient time,
        For an ensemble of initial conditions, can do the same transient time for each, or the
        `convergence_time` for the first one, and a smaller `reconvergence_time` for the subsequent ones.
        This results into initial conditions on the attractor, stored in :attr:`ic`.

        Parameters
        ----------
        convergence_time: float
            Transient time needed to converge to the attractor.
        dt: float
            Timestep of the transient integration.
        pert_size:float, optional
            If the reconvergence is activated, size of the perturbation to add to the previous ic to find
            the next one. Default to 0.01 .
        reconvergence_time: None or float, optional
            Transient time for the subsequent trajectories after the first long `transient_time`.
        forward: bool, optional
            Whether to integrate the ODEs forward or backward in time. In case of backward integration, the
            initial condition `ic` becomes a final condition. Default to forward integration.
        number_of_trajectories: int
            Number of initial conditions to find. Default to 1.  Inactive if `ic` is provided.
        ic: None or ~numpy.ndarray(float), optional
            Initial (or final) conditions of the system. Can be a 1D or a 2D array:

            * 1D: Provide a single initial condition.
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Provide an ensemble of initial condition.
              Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`,
              and where `n_traj` is the number of initial conditions.

            If `None`, use `number_trajectories` random initial conditions. Default to `None`.
            If the `forward` argument is `False`, it specifies final conditions.
        reconverge: bool
            Use or not the smaller transient time reconvergence with a perturbation
            after the first initial conditions have been computed. If activated, only use the :attr:`num_threads`
            first initial conditions of the `ic` arguments. Default to `False`.
        """

        if reconverge is None:
            reconverge = False

        if ic is None:
            if self.n_dim is not None:
                i = self.n_dim
            else:
                i = 1
                while True:
                    self.ic = np.zeros(i)
                    try:
                        x = self.func(0., self.ic)
                    except:
                        i += 1
                    else:
                        break

                i = len(self.func(0., self.ic))

            if number_of_trajectories > self.num_threads:
                reconverge = True
                tmp_ic = np.zeros((number_of_trajectories, i))
                tmp_ic[:self.num_threads] = np.random.randn(self.num_threads, i)
            else:
                tmp_ic = np.random.randn(number_of_trajectories, i)
        else:
            tmp_ic = ic.copy()
            if len(tmp_ic.shape) > 1:
                number_of_trajectories = tmp_ic.shape[0]

        if reconverge and reconvergence_time is not None:

            self.integrate(0., convergence_time, dt, ic=tmp_ic[:self.num_threads], write_steps=0, forward=forward)
            t, x = self.get_trajectories()
            tmp_ic[:self.num_threads] = x
            if number_of_trajectories - self.num_threads > self.num_threads:
                next_len = self.num_threads
            else:
                next_len = number_of_trajectories - self.num_threads

            index = self.num_threads
            while True:
                perturbation = pert_size * np.random.randn(next_len, x.shape[1])
                self.integrate(0., reconvergence_time, dt, ic=x[:next_len]+perturbation, write_steps=0, forward=forward)
                t, x = self.get_trajectories()
                tmp_ic[index:index+next_len] = x
                index += next_len
                if number_of_trajectories - index > self.num_threads:
                    next_len = self.num_threads
                else:
                    next_len = number_of_trajectories - index
                if next_len <= 0:
                    break
            self.ic = tmp_ic
        else:
            self.integrate(0., convergence_time, dt, ic=tmp_ic, write_steps=0, forward=forward)
            t, x = self.get_trajectories()
            self.ic = x

    def integrate(self, t0, t, dt, ic=None, forward=True, write_steps=1):
        """Integrate the ordinary differential equations (ODEs)

        .. math:: \\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

        with a specified `Runge-Kutta method`_ and workers. The function :math:`\\boldsymbol{f}` is the `Numba`_ jitted
        function stored in :attr:`func`. The result of the integration can be obtained afterward by calling
        :meth:`get_trajectories`.

        .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
        .. _Numba: https://numba.pydata.org/

        Parameters
        ----------
        t0: float
            Initial time of the time integration. Corresponds to the initial condition's `ic` time.
            Important if the ODEs are non-autonomous.
        t: float
            Final time of the time integration. Corresponds to the final condition.
            Important if the ODEs are non-autonomous.
        dt: float
            Timestep of the integration.
        ic: None or ~numpy.ndarray(float), optional
            Initial (or final) conditions of the system. Can be a 1D or a 2D array:

            * 1D: Provide a single initial condition.
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Provide an ensemble of initial condition.
              Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`,
              and where `n_traj` is the number of initial conditions.

            If `None`, use the initial conditions stored in :attr:`ic`.
            If then :attr:`ic` is `None`, use a zero initial condition.
            Default to `None`.
            If the `forward` argument is `False`, it specifies final conditions.
        forward: bool, optional
            Whether to integrate the ODEs forward or backward in time. In case of backward integration, the
            initial condition `ic` becomes a final condition. Default to forward integration.
        write_steps: int, optional
            Save the state of the integration in memory every `write_steps` steps. The other intermediary
            steps are lost. It determines the size of the returned objects. Default is 1.
            Set to 0 to return only the final state.
        """
        if self.func is None:
            print('No function to integrate defined!')
            return 0

        if ic is None:
            if self.ic is None:
                if self.n_dim is not None:
                    i = self.n_dim
                else:
                    i = 1
                    while True:
                        self.ic = np.zeros(i)
                        try:
                            x = self.func(0., self.ic)
                        except:
                            i += 1
                        else:
                            break

                    i = len(self.func(0., self.ic))

                self.ic = np.zeros(i)
        else:
            self.ic = ic

        if len(self.ic.shape) == 1:
            self.ic = self.ic.reshape((1, -1))

        self.n_traj = self.ic.shape[0]
        self.n_dim = self.ic.shape[1]
        self._time = np.concatenate((np.arange(t0, t, dt), np.full((1,), t)))
        self._write_steps = write_steps

        if forward:
            self._time_direction = 1
        else:
            self._time_direction = -1

        if write_steps == 0:
            self.n_records = 1
        else:
            tot = self._time[::self._write_steps]
            self.n_records = len(tot)
            if tot[-1] != self._time[-1]:
                self.n_records += 1

        self._recorded_traj = np.zeros((self.n_traj, self.n_dim, self.n_records))

        for i in range(self.n_traj):
            self._ics_queue.put((i, self._time, self.ic[i], self._time_direction, self._write_steps))

        self._ics_queue.join()

        for i in range(self.n_traj):
            args = self._traj_queue.get()
            self._recorded_traj[args[0]] = args[1]

    def get_trajectories(self):
        """Returns the result of the previous integrator integration.

        Returns
        -------
        time, traj: ~numpy.ndarray
            The result of the integration:

            * **time:** Time at which the state of the system was saved. Array of shape (:attr:`n_records`,).
            * **traj:** Saved dynamical system states. 3D array of shape (:attr:`n_traj`, :attr:`n_dim`, :attr:`n_records`).
              If :attr:`n_traj` = 1, a 2D array of shape (:attr:`n_dim`, :attr:`n_records`) is returned instead.
        """
        if self._write_steps > 0:
            if self._time_direction == 1:
                if self._time[::self._write_steps][-1] == self._time[-1]:
                    return self._time[::self._write_steps], np.squeeze(self._recorded_traj)
                else:
                    return np.concatenate((self._time[::self._write_steps], np.full((1,), self._time[-1]))), \
                           np.squeeze(self._recorded_traj)
            else:
                rtime = reverse(self._time[::-self._write_steps])
                if rtime[0] == self._time[0]:
                    return rtime, np.squeeze(self._recorded_traj)
                else:
                    return np.concatenate((np.full((1,), self._time[0]), rtime)), np.squeeze(self._recorded_traj)

        else:
            return self._time[-1], np.squeeze(self._recorded_traj)

    def get_ic(self):
        """Returns the initial conditions stored in the integrator.

        Returns
        -------
        ~numpy.ndarray
            The initial conditions.
        """
        return self.ic

    def set_ic(self, ic):
        """Direct setter for the integrator's initial conditions

        Parameters
        ----------
        ic: ~numpy.ndarray(float)
            Initial condition of the system. Can be a 1D or a 2D array:

            * 1D: Provide a single initial condition.
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Provide an ensemble of initial condition.
              Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`,
              and where `n_traj` is the number of initial conditions.
        """
        self.ic = ic


class TrajectoryProcess(multiprocessing.Process):
    """:class:`RungeKuttaIntegrator`'s workers class. Allows to multi-thread time integration.

    .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
    .. _Numba: https://numba.pydata.org/

    Parameters
    ----------
    processID: int
        Number identifying the worker.
    func: callable
        `Numba`_-jitted function to integrate assigned to the worker.
    b: ~numpy.ndarray, optional
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
    c: ~numpy.ndarray, optional
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
    a: ~numpy.ndarray, optional
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
    ics_queue: multiprocessing.JoinableQueue
        Queue to which the worker ask for initial conditions and parameters input.
    traj_queue: multiprocessing.Queue
        Queue to which the worker returns the integration results.

    Attributes
    ----------
    processID: int
        Number identifying the worker.
    func: callable
        `Numba`_-jitted function to integrate assigned to the worker.
    b: ~numpy.ndarray
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
    c: ~numpy.ndarray
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
    a: ~numpy.ndarray
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
    """
    def __init__(self, processID, func, b, c, a, ics_queue, traj_queue):

        super().__init__()
        self.processID = processID
        self._ics_queue = ics_queue
        self._traj_queue = traj_queue
        self.func = func
        self.a = a
        self.b = b
        self.c = c

    def run(self):
        """Main worker computing routine. Perform the time integration with the fetched initial conditions and parameters."""

        while True:

            args = self._ics_queue.get()

            recorded_traj = _integrate_runge_kutta_jit(self.func, args[1], args[2][np.newaxis, :], args[3], args[4],
                                                       self.b, self.c, self.a)

            self._traj_queue.put((args[0], recorded_traj))

            self._ics_queue.task_done()


class RungeKuttaTglsIntegrator(object):
    """Class to integrate simultaneously the ordinary differential equations (ODEs)

    .. math:: \dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

    and its tangent linear model, i.e. the linearized ODEs

    .. math :: \dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}

    where :math:`\\boldsymbol{\mathrm{J}} = \\frac{\partial \\boldsymbol{f}}{\partial \\boldsymbol{x}}` is the
    Jacobian matrix of :math:`\\boldsymbol{f}`, with a specified `Runge-Kutta method`_.
    To solve this equation, one has to integrate simultaneously both ODEs. This class does so with a set
    of :class:`TglsTrajectoryProcess` workers.

    The function :math:`\\boldsymbol{f}` and :math:`\\boldsymbol{J}` should
    be `Numba`_ jitted functions. These functions must have a signature ``f(t, x)`` and ``J(t, x)`` where ``x`` is
    the state value and ``t`` is the time.

    .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
    .. _Numba: https://numba.pydata.org/
    .. _fundamental matrix of solutions: https://en.wikipedia.org/wiki/Fundamental_matrix_(linear_differential_equation)

    Parameters
    ----------
    num_threads: None or int, optional
        Number of :class:`TrajectoryProcess` workers (threads) to use. If `None`, use the number of machine's
        cores available. Default to `None`.
    b: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    c: None or ~numpy.ndarray, optional
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    a: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    number_of_dimensions: None or int, optional
        Allow to hardcode the dynamical system dimension. If `None`, evaluate the dimension from the
        callable :attr:`func`. Default to `None`.

    Attributes
    ----------
    num_threads: int
        Number of :class:`TrajectoryProcess` workers (threads) to use.
    b: ~numpy.ndarray
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
    c: ~numpy.ndarray
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
    a: ~numpy.ndarray
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
    n_dim: int
        Dynamical system dimension.
    n_traj: int
        The number of trajectories (initial conditions) of the non-linear ODEs computed at the last integration
        performed by the integrator.
    n_tg_traj: int
        The number of trajectories (initial conditions) the linear ODEs computed at the last integration
        performed by the integrator.
    n_records: int
        The number of saved states of the last integration performed by the integrator.
    ic: ~numpy.ndarray
        Store the integrator non-linear ODEs initial conditions.
    tg_ic: ~numpy.ndarray
        Store the integrator linear ODEs initial conditions.
    func: callable
        Last function :math:`\\boldsymbol{f}` used by the integrator to integrate.
    func_jac: callable
        Last Jacobian matrix function :math:`\\boldsymbol{J}` used by the integrator to integrate.
    """
    def __init__(self, num_threads=None, b=None, c=None, a=None, number_of_dimensions=None):

        if num_threads is None:
            self.num_threads = multiprocessing.cpu_count()
        else:
            self.num_threads = num_threads

        # Default is RK4
        if a is None and b is None and c is None:
            self.c = np.array([0., 0.5, 0.5, 1.])
            self.b = np.array([1./6, 1./3, 1./3, 1./6])
            self.a = np.zeros((len(self.c), len(self.b)))
            self.a[1, 0] = 0.5
            self.a[2, 1] = 0.5
            self.a[3, 2] = 1.
        else:
            self.a = a
            self.b = b
            self.c = c

        self.ic = None
        self.tg_ic = None
        self._time = None
        self._recorded_traj = None
        self._recorded_fmatrix = None
        self.n_traj = 0
        self.n_tg_traj = 0
        self.n_dim = number_of_dimensions
        self.n_records = 0
        self._write_steps = 0

        self._time_direction = 1
        self._adjoint = False
        self._boundary = None
        self._inverse = 1.

        self.func = None
        self.func_jac = None

        self._ics_queue = None
        self._traj_queue = None

        self._processes_list = list()

    def terminate(self):
        """Stop the workers (threads) and release the resources of the integrator."""

        for process in self._processes_list:

            process.terminate()
            process.join()

    def start(self):
        """Start or restart the workers (threads) of the integrator.

        Warnings
        --------
        If the integrator was not previously terminated, it will be terminated first in the case
        of a restart.
        """

        self.terminate()

        self._processes_list = list()
        self._ics_queue = multiprocessing.JoinableQueue()
        self._traj_queue = multiprocessing.Queue()

        for i in range(self.num_threads):
            self._processes_list.append(TglsTrajectoryProcess(i, self.func, self.func_jac, self.b, self.c, self.a,
                                                              self._ics_queue, self._traj_queue))

        for process in self._processes_list:
            process.daemon = True
            process.start()

    def set_func(self, f, fjac, ic_init=True):
        """Set the `Numba`_-jitted function :math:`\\boldsymbol{f}` and Jacobian matrix function
        :math:`\\boldsymbol{\\mathrm{J}}` to integrate.

        .. _Numba: https://numba.pydata.org/

        Parameters
        ----------
        f: callable
            The `Numba`_-jitted function :math:`\\boldsymbol{f}`.
            Should have the signature ``f(t, x)`` where ``x`` is the state value and ``t`` is the time.
        fjac: callable
            The `Numba`_-jitted Jacobian matrix function :math:`\\boldsymbol{J}`.
            Should have the signature ``J(t, x)`` where ``x`` is the state value and ``t`` is the time.
        ic_init: bool, optional
            Re-initialize or not the initial conditions of the integrator. Default to `True`.

        Warnings
        --------
        This function restarts the integrator!
        """

        self.func = f
        self.func_jac = fjac
        if ic_init:
            self.ic = None
        self.start()

    def set_bca(self, b=None, c=None, a=None, ic_init=True):
        """Set the coefficients of the `Runge-Kutta method`_ and restart the integrator. s

        .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

        Parameters
        ----------
        b: None or ~numpy.ndarray, optional
            Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
            If `None`, does not reinitialize these coefficients.
        c: None or ~numpy.ndarray, optional
            Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
            If `None`, does not reinitialize these coefficients.
        a: None or ~numpy.ndarray, optional
            Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
            If `None`, does not reinitialize these coefficients.
        ic_init: bool, optional
            Re-initialize or not the initial conditions of the integrator. Default to `True`.

        """

        if a is not None:
            self.a = a
        if b is not None:
            self.b = b
        if c is not None:
            self.c = c
        if ic_init:
            self.ic = None
        self.start()

    def initialize(self, convergence_time, dt, pert_size=0.01, reconvergence_time=None, forward=True,
                   number_of_trajectories=1, ic=None, reconverge=None):
        """Initialize the integration on an attractor by running it for a transient time,
        For an ensemble of initial conditions, can do the same transient time for each, or the
        `convergence_time` for the first one, and a smaller `reconvergence_time` for the subsequent ones.
        This results into initial conditions on the attractor, stored in :attr:`ic`.

        Parameters
        ----------
        convergence_time: float
            Transient time needed to converge to the attractor.
        dt: float
            Timestep of the transient integration.
        pert_size:float, optional
            If the reconvergence is activated, size of the perturbation to add to the previous ic to find
            the next one. Default to 0.01 .
        reconvergence_time: None or float, optional
            Transient time for the subsequent trajectories after the first long `transient_time`.
        forward: bool, optional
            If true, integrate the ODEs forward in time, else, integrate backward in time. In case of backward integration, the
            initial condition `ic` becomes a final condition. Default to forward integration.
        number_of_trajectories: int
            Number of initial conditions to find. Default to 1.  Inactive if `ic` is provided.
        ic: None or ~numpy.ndarray(float), optional
            Initial condition of the system. Can be a 1D or a 2D array:

            * 1D: Provide a single initial condition.
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Provide an ensemble of initial condition.
              Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`,
              and where `n_traj` is the number of initial conditions.

            If `None`, use `number_trajectories` random initial conditions. Default to `None`.
        reconverge: bool
            Use or not the smaller transient time reconvergence with a perturbation
            after the first initial conditions have been computed. If activated, only use the :attr:`num_threads`
            first initial conditions of the `ic` arguments. Default to `False`.
        """

        if reconverge is None:
            reconverge = False

        if ic is None:
            if self.n_dim is not None:
                i = self.n_dim
            else:
                i = 1
                while True:
                    self.ic = np.zeros(i)
                    try:
                        x = self.func(0., self.ic)
                    except:
                        i += 1
                    else:
                        break

                i = len(self.func(0., self.ic))

            if number_of_trajectories > self.num_threads:
                reconverge = True
                tmp_ic = np.zeros((number_of_trajectories, i))
                tmp_ic[:self.num_threads] = np.random.randn(self.num_threads, i)
            else:
                tmp_ic = np.random.randn(number_of_trajectories, i)
        else:
            tmp_ic = ic.copy()
            if len(tmp_ic.shape) > 1:
                number_of_trajectories = tmp_ic.shape[0]

        if reconverge and reconvergence_time is not None:

            self.integrate(0., convergence_time, dt, ic=tmp_ic[:self.num_threads], write_steps=0, forward=forward)
            t, x, fm = self.get_trajectories()
            tmp_ic[:self.num_threads] = x
            if number_of_trajectories - self.num_threads > self.num_threads:
                next_len = self.num_threads
            else:
                next_len = number_of_trajectories - self.num_threads

            index = self.num_threads
            while True:
                perturbation = pert_size * np.random.randn(next_len, x.shape[1])
                self.integrate(0., reconvergence_time, dt, ic=x[:next_len]+perturbation, write_steps=0, forward=forward)
                t, x, fm = self.get_trajectories()
                tmp_ic[index:index+next_len] = x
                index += next_len
                if number_of_trajectories - index > self.num_threads:
                    next_len = self.num_threads
                else:
                    next_len = number_of_trajectories - index
                if next_len <= 0:
                    break
            self.ic = tmp_ic
        else:
            self.integrate(0., convergence_time, dt, ic=tmp_ic, write_steps=0, forward=forward)
            t, x, fm = self.get_trajectories()
            self.ic = x

    def integrate(self, t0, t, dt, ic=None, tg_ic=None, forward=True, adjoint=False, inverse=False, boundary=None,
                  write_steps=1):
        """Integrate simultaneously the non-linear and linearized ordinary differential equations (ODEs)

        .. math:: \\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

        and

        .. math :: \\dot{\\boldsymbol{\\delta x}} = \\boldsymbol{\\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\\delta x}

        with a specified `Runge-Kutta method`_ and workers.
        The function :math:`\\boldsymbol{f}` is the `Numba`_ jitted function stored in :attr:`func`.
        The function :math:`\\boldsymbol{J}` is the `Numba`_ jitted function stored in :attr:`func_jac`.
        The result of the integration can be obtained afterward by calling :meth:`get_trajectories`.

        .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
        .. _Numba: https://numba.pydata.org/
        .. _fundamental matrix of solutions: https://en.wikipedia.org/wiki/Fundamental_matrix_(linear_differential_equation)

        Parameters
        ----------
        t0: float
            Initial time of the time integration. Corresponds to the initial condition.
        t: float
            Final time of the time integration. Corresponds to the final condition.
        dt: float
            Timestep of the integration.
        ic: None or ~numpy.ndarray(float), optional
            Initial (or final) conditions of the system. Can be a 1D or a 2D array:

            * 1D: Provide a single initial condition.
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Provide an ensemble of initial condition.
              Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`,
              and where `n_traj` is the number of initial conditions.

            If `None`, use the initial conditions stored in :attr:`ic`.
            If then :attr:`ic` is `None`, use a zero initial condition.
            Default to `None`.
            If the `forward` argument is `False`, it specifies final conditions.
        tg_ic: None or ~numpy.ndarray(float), optional
            Initial (or final) conditions of the linear ODEs
            :math:`\\dot{\\boldsymbol{\\delta x}} = \\boldsymbol{\\mathrm{J}}(t, \\boldsymbol{x}) \\cdot \\boldsymbol{\\delta x}`. \n
            Can be a 1D, a 2D or a 3D array:

            * 1D: Provide a single initial condition. This initial condition of the linear ODEs will be the same used for each
              initial condition `ic` of the ODEs :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Two sub-cases:

                + If `tg_ic.shape[0]`=`ic.shape[0]`, assumes that each initial condition `ic[i]` of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`,
                  correspond to a different initial condition `tg_ic[i]`.
                + Else, assumes and integrate an ensemble of `n_tg_traj` initial condition of the linear ODEs for each
                  initial condition of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`.
            * 3D: An array of shape (`n_traj`, `n_tg_traj`, `n_dim`) which provide an ensemble of `n_tg_ic` initial conditions
              specific to each of the `n_traj` initial conditions of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`.

            If `None`, use the identity matrix as initial condition, returning the `fundamental matrix of solutions`_ of the
            linear ODEs.
            Default to `None`.
            If the `forward` argument is `False`, it specifies final conditions.

        forward: bool, optional
            If true, integrate the ODEs forward in time, else, integrate backward in time. In case of backward integration, the
            initial condition `ic` becomes a final condition. Default to forward integration.
        adjoint: bool, optional
            If true, integrate the tangent :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}` ,
            else, integrate the adjoint linear model :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}^T(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}`.
            Integrate the tangent model by default.
        inverse: bool, optional
            Wheter or not to invert the Jacobian matrix
            :math:`\\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \\rightarrow \\boldsymbol{\mathrm{J}}^{-1}(t, \\boldsymbol{x})`.
            `False` by default.
        boundary: None or callable, optional
            Allow to add a inhomogeneous term to linear ODEs:
            :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x} + \Psi(t, \\boldsymbol{x})`.
            The boundary :math:`\Psi` should have the same signature as :math:`\\boldsymbol{\mathrm{J}}`, i.e.  ``func(t, x)``.
            If `None`, don't add anything (homogeneous case). `None` by default.
        write_steps: int, optional
            Save the state of the integration in memory every `write_steps` steps. The other intermediary
            steps are lost. It determines the size of the returned objects. Default is 1.
            Set to 0 to return only the final state.
        """

        if self.func is None or self.func_jac is None:
            print('No function to integrate defined!')
            return 0

        if ic is None:
            if self.ic is None:
                if self.n_dim is not None:
                    i = self.n_dim
                else:
                    i = 1
                    while True:
                        self.ic = np.zeros(i)
                        try:
                            x = self.func(0., self.ic)
                        except:
                            i += 1
                        else:
                            break

                    i = len(self.func(0., self.ic))

                self.ic = np.zeros(i)
        else:
            self.ic = ic

        if len(self.ic.shape) == 1:
            self.ic = self.ic.reshape((1, -1))

        self.n_traj = self.ic.shape[0]
        self.n_dim = self.ic.shape[1]
        self._time = np.concatenate((np.arange(t0, t, dt), np.full((1,), t)))
        self._write_steps = write_steps

        if tg_ic is None:
            tg_ic = np.eye(self.ic.shape[1])

        tg_ic_sav = tg_ic.copy()

        if len(tg_ic.shape) == 1:
            tg_ic = tg_ic.reshape((1, -1, 1))
            ict = tg_ic.copy()
            for i in range(self.n_traj-1):
                ict = np.concatenate((ict, tg_ic))
            self.tg_ic = ict
        elif len(tg_ic.shape) == 2:
            if tg_ic.shape[0] == self.n_traj:
                self.tg_ic = tg_ic[..., np.newaxis]
            else:
                tg_ic = tg_ic[np.newaxis, ...]
                tg_ic = np.swapaxes(tg_ic, 1, 2)
                ict = tg_ic.copy()
                for i in range(self.n_traj-1):
                    ict = np.concatenate((ict, tg_ic))
                self.tg_ic = ict
        elif len(tg_ic.shape) == 3:
            if tg_ic.shape[1] != self.n_dim:
                self.tg_ic = np.swapaxes(tg_ic, 1, 2)

        self.n_tg_traj = self.tg_ic.shape[1]

        if forward:
            self._time_direction = 1
        else:
            self._time_direction = -1

        self._adjoint = adjoint

        if boundary is None:
            self._boundary = _zeros_func
        else:
            self._boundary = boundary

        self._inverse = 1.
        if inverse:
            self._inverse *= -1.

        if write_steps == 0:
            self.n_records = 1
        else:
            tot = self._time[::self._write_steps]
            self.n_records = len(tot)
            if tot[-1] != self._time[-1]:
                self.n_records += 1

        self._recorded_traj = np.zeros((self.n_traj, self.n_dim, self.n_records))
        self._recorded_fmatrix = np.zeros((self.n_traj, self.tg_ic.shape[1], self.tg_ic.shape[2], self.n_records))

        for i in range(self.n_traj):
            self._ics_queue.put((i, self._time, self.ic[i], self.tg_ic[i], self._time_direction, self._write_steps,
                                 self._adjoint, self._inverse, self._boundary))

        self._ics_queue.join()

        for i in range(self.n_traj):
            args = self._traj_queue.get()
            self._recorded_traj[args[0]] = args[1]
            self._recorded_fmatrix[args[0]] = args[2]

        if len(tg_ic_sav.shape) == 2:
            if self._recorded_fmatrix.shape[1:3] != tg_ic_sav.shape:
                self._recorded_fmatrix = np.swapaxes(self._recorded_fmatrix, 1, 2)

        elif len(tg_ic_sav.shape) == 3:
            if tg_ic_sav.shape[1] != self.n_dim:
                if self._recorded_fmatrix.shape[:3] != tg_ic_sav.shape:
                    self._recorded_fmatrix = np.swapaxes(self._recorded_fmatrix, 1, 2)

    def get_trajectories(self):
        """Returns the result of the previous integrator integration.

        Returns
        -------
        time, traj, tg_traj: ~numpy.ndarray
            The result of the integration:

            * **time:** time at which the state of the system was saved. Array of shape (:attr:`n_records`,).
            * **traj:** Saved dynamical system states. 3D array of shape (:attr:`n_traj`, :attr:`n_dim`, :attr:`n_records`).
              If :attr:`n_traj` = 1, a 2D array of shape (:attr:`n_dim`, :attr:`n_records`) is returned instead.
            * **tg_traj:** Saved states of the linear ODEs.
              Depending on the input initial conditions of both ODEs,
              it is at maximum a 4D array of shape
              (:attr:`n_traj`, :attr:`n_tg_traj`, :attr:`n_dim`, :attr:`n_records`).
              If one of the dimension is 1, it is squeezed.
        """
        if self._write_steps > 0:
            if self._time_direction == 1:
                if self._time[::self._write_steps][-1] == self._time[-1]:
                    return self._time[::self._write_steps], np.squeeze(self._recorded_traj), \
                           np.squeeze(self._recorded_fmatrix)
                else:
                    return np.concatenate((self._time[::self._write_steps], np.full((1,), self._time[-1]))), \
                           np.squeeze(self._recorded_traj), np.squeeze(self._recorded_fmatrix)
            else:
                rtime = reverse(self._time[::-self._write_steps])
                if rtime[0] == self._time[0]:
                    return rtime, np.squeeze(self._recorded_traj), np.squeeze(self._recorded_fmatrix)
                else:
                    return np.concatenate((np.full((1,), self._time[0]), rtime)), np.squeeze(self._recorded_traj), \
                           np.squeeze(self._recorded_fmatrix)
        else:
            return self._time[-1], np.squeeze(self._recorded_traj), np.squeeze(self._recorded_fmatrix)

    def get_ic(self):
        """Returns the initial conditions of the non-linear ODEs stored in the integrator.

        Returns
        -------
        ~numpy.ndarray
            The initial conditions.
        """
        return self.ic

    def set_ic(self, ic):
        """Direct setter for the integrator's non-linear ODEs initial conditions

        Parameters
        ----------
        ic: ~numpy.ndarray(float)
            Initial condition of the system. Can be a 1D or a 2D array:

            * 1D: Provide a single initial condition.
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Provide an ensemble of initial condition.
              Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`,
              and where `n_traj` is the number of initial conditions.
        """
        self.ic = ic

    def get_tg_ic(self):
        """Returns the initial conditions of the linear ODEs stored in the integrator.

        Returns
        -------
        ~numpy.ndarray
            The initial conditions.
        """
        return self.tg_ic

    def set_tg_ic(self, tg_ic):
        """Direct setter for the integrator's linear ODEs initial conditions

        Parameters
        ----------
        tg_ic: ~numpy.ndarray(float)
            Initial condition of the linear ODEs
            :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}`. \n
            Can be a 1D, a 2D or a 3D array:

            * 1D: Provide a single initial condition. This initial condition of the linear ODEs will be the same used for each
              initial condition `ic` of the ODEs :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`
              Should be of shape (`n_dim`,) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`.
            * 2D: Two sub-cases:

                + If `tg_ic.shape[0]`=`ic.shape[0]`, assumes that each initial condition `ic[i]` of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`,
                  correspond to a different initial condition `tg_ic[i]`.
                + Else, assumes and integrate an ensemble of `n_tg_traj` initial condition of the linear ODEs for each
                  initial condition of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`.
            * 3D: An array of shape (`n_traj`, `n_tg_traj`, `n_dim`) which provide an ensemble of `n_tg_ic` initial conditions
              specific to each of the `n_traj` initial conditions of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`.
        """
        self.tg_ic = tg_ic


class TglsTrajectoryProcess(multiprocessing.Process):
    """:class:`RungeKuttaTglsIntegrator`'s workers class. Allows to multi-thread time integration.

    .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
    .. _Numba: https://numba.pydata.org/

    Parameters
    ----------
    processID: int
        Number identifying the worker.
    func: callable
        `Numba`_-jitted function to integrate assigned to the worker.
    func_jac: callable
        `Numba`_-jitted Jacobian matrix function to integrate assigned to the worker.
    b: ~numpy.ndarray, optional
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
    c: ~numpy.ndarray, optional
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
    a: ~numpy.ndarray, optional
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
    ics_queue: multiprocessing.JoinableQueue
        Queue to which the worker ask for initial conditions input.
    traj_queue: multiprocessing.Queue
        Queue to which the worker returns the integration results.

    Attributes
    ----------
    processID: int
        Number identifying the worker.
    func: callable
        `Numba`_-jitted function to integrate assigned to the worker.
    func_jac: callable
        `Numba`_-jitted Jacobian matrix function to integrate assigned to the worker.
    b: ~numpy.ndarray
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
    c: ~numpy.ndarray
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
    a: ~numpy.ndarray
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
    """
    def __init__(self, processID, func, func_jac, b, c, a, ics_queue, traj_queue):

        super().__init__()
        self.processID = processID
        self._ics_queue = ics_queue
        self._traj_queue = traj_queue
        self.func = func
        self.func_jac = func_jac
        self.a = a
        self.b = b
        self.c = c

    def run(self):
        """Main worker computing routine. Perform the time integration with the fetched initial conditions and parameters."""

        while True:

            args = self._ics_queue.get()

            recorded_traj, recorded_fmatrix = _integrate_runge_kutta_tgls_jit(self.func, self.func_jac, args[1], args[2][np.newaxis, ...],
                                                                              args[3][np.newaxis, ...], args[4], args[5],
                                                                              self.b, self.c, self.a,
                                                                              args[6], args[7], args[8])

            self._traj_queue.put((args[0], recorded_traj, recorded_fmatrix))

            self._ics_queue.task_done()


if __name__ == "__main__":

    import matplotlib.pyplot as plt
    from scipy.integrate import odeint

    @njit
    def f(t, x):
        return - np.array([1., 2., 3.]) * x

    def fr(x, t):
        return f(t, x)

    ic = np.random.randn(6).reshape(2, 3)

    integrator = RungeKuttaIntegrator()

    integrator.set_func(f)
    integrator.integrate(0., 10., 0.01, ic=ic, write_steps=1)

    time, r = integrator.get_trajectories()

    t = np.arange(0., 10., 0.1)
    t = np.concatenate((t[::3], np.full((1,), 10.)))
    rl = list()
    for i in range(ic.shape[0]):
        rl.append(odeint(fr, ic[i], t).T)

    plt.figure()
    for i in range(ic.shape[0]):
        p, = plt.plot(time, r[i, 0])
        c = p.get_color()
        plt.plot(t, rl[i][0], color=c, ls='--')
        for j in range(1, ic.shape[1]):
            p, = plt.plot(time, r[i, j], color=c)
            plt.plot(t, rl[i][j], color=c, ls='--')
    plt.title('Forward')

    integrator.integrate(0., 10., 0.01, ic=ic, forward=False, write_steps=1)
    timet, rt = integrator.get_trajectories()
    rlt = list()
    for i in range(ic.shape[0]):
        rlt.append(odeint(fr, ic[i], reverse(t)).T)

    plt.figure()
    for i in range(ic.shape[0]):
        p, = plt.plot(timet, rt[i, 0])
        c = p.get_color()
        plt.plot(t, reverse(rlt[i][0]), color=c, ls='--')
        for j in range(1, ic.shape[1]):
            p, = plt.plot(timet, rt[i, j], color=c)
            plt.plot(t, reverse(rlt[i][j]), color=c, ls='--')
    plt.title('Backward')

    integrator.integrate(0., 10., 0.01, ic=ic, write_steps=0)
    tt, re = integrator.get_trajectories()
    print(tt)
    print(r[0, :, -1], re[0])
    plt.show(block=False)

    a = 0.25
    F = 16.
    G = 3.
    b = 6.


    @njit
    def fL84(t, x):
        xx = -x[1] ** 2 - x[2] ** 2 - a * x[0] + a * F
        yy = x[0] * x[1] - b * x[0] * x[2] - x[1] + G
        zz = b * x[0] * x[1] + x[0] * x[2] - x[2]
        return np.array([xx, yy, zz])


    @njit
    def DfL84(t, x):
        return np.array([[     -a        , -2. * x[1], -2. * x[2]],
                         [x[1] - b * x[2], -1. + x[0], -b * x[0]],
                         [b * x[1] + x[2],  b * x[0], -1. + x[0]]])


    integrator.set_func(fL84)
    integrator.integrate(0., 10000., 0.01, write_steps=0)
    tt, traj = integrator.get_trajectories()
    integrator.integrate(0.,20.,0.01,ic=traj, write_steps=10)
    tt, traj = integrator.get_trajectories()
    integrator.integrate(0.,20.,0.01,ic=traj[:,-1],write_steps=10, forward=False)
    ttb, trajb = integrator.get_trajectories()

    plt.figure()
    plt.plot(tt, traj[0])
    plt.plot(ttb, trajb[0])
    plt.show(block=False)

    plt.title('Lorenz 63 - Forward then backward')

    integrator.integrate(0., 10000., 0.01, write_steps=0)
    tt, traj = integrator.get_trajectories()
    integrator.integrate(0.,20.,0.01,ic=traj,write_steps=10, forward=False)
    tt, trajb = integrator.get_trajectories()
    integrator.integrate(0.,20.,0.01,ic=trajb[:,0],write_steps=10)
    ttb, traj = integrator.get_trajectories()

    plt.figure()
    plt.plot(tt, traj[0])
    plt.plot(ttb, trajb[0])
    plt.show(block=False)

    plt.title('Lorenz 63 - Backward then forward')
    integrator.terminate()

    tgls_integrator = RungeKuttaTglsIntegrator()
    tgls_integrator.set_func(fL84, DfL84)

    @njit
    def tboundary(t, x):
        return np.array([0.,x[1],0.])

    ic = np.random.randn(4, 3)
    tgls_integrator.initialize(10., 0.01, ic=ic)
    tgls_integrator.integrate(0., 20., 0.01, write_steps=10, tg_ic=np.zeros(3), boundary=tboundary)
    # x, fm = _integrate_runge_kutta_tgls_jit(fL84, DfL84, tgls_integrator.time, tgls_integrator.ic[0][np .newaxis,...], np.zeros((1,3,1)), 1, 1, tgls_integrator.b, tgls_integrator.c, tgls_integrator.a, False, 1., tboundary)

    t, x, fm = tgls_integrator.get_trajectories()

    tgls_integrator.terminate()
